import datetime
import pytest

from qarnot.status import Status

from .mock_status import default_json_status

class TestStatusProperties:
    @pytest.mark.parametrize("property_name, expected_value", [
        ("last_update_timestamp", default_json_status["lastUpdateTimestamp"]),
    ])
    def test_create_status_hydrate_property_values(self, property_name,  expected_value):
        status = Status(default_json_status)
        assert getattr(status, property_name) == expected_value

    @pytest.mark.parametrize("property_names, expected_value", [
        (["running_instances_info", "snapshot_results"], default_json_status["runningInstancesInfo"]["snapshotResults"]),
        (["running_instances_info", "running_core_count_by_cpu_model"], default_json_status["runningInstancesInfo"]["runningCoreCountByCpuModel"]),
        (["execution_time_by_cpu_model", 0, "model"], default_json_status["executionTimeByCpuModel"][0]["model"]),
        (["execution_time_by_cpu_model", 0, "time"], default_json_status["executionTimeByCpuModel"][0]["time"]),
        (["execution_time_by_cpu_model", 0, "core"], default_json_status["executionTimeByCpuModel"][0]["core"]),
    ])
    def test_create_status_hydrate_subproperty_values(self, property_names,  expected_value):
        status = Status(default_json_status)
        value = getattr(status, property_names[0])
        for property_name in property_names[1:]:
            if type(property_name) is int:
                value = value[property_name]
            else:
                value = getattr(value, property_name)
        assert value == expected_value



"""Rest API for submitting qarnot jobs in Python."""


# Copyright 2017 Qarnot computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .exceptions import QarnotGenericException

__all__ = ["task", "connection", "bucket", "pool",
           "storage", "status", "job", "advanced_bucket"]


def raise_on_error(response):
    if response.status_code == 503:
        raise QarnotGenericException("Service Unavailable")
    if response.status_code < 200 or response.status_code >= 300:
        try:
            raise QarnotGenericException(response.json()['message'])
        except ValueError as value:
            raise QarnotGenericException(response.text) from value


def get_url(key, **kwargs):
    """Get and format the url for the given key.
    """
    urls = {
        'jobs': u'/jobs/',  # POST -> Submit Job
        'paginate jobs': u'/jobs/paginate',  # GET -> paginate jobs
        'job update': u'/jobs/{uuid}',  # GET -> result; DELETE -> abort
        'job delete': u'/jobs/{uuid}?force={force}',  # DELETE -> delete job
        'jobs search': u'/jobs/search',  # POST -> make a custom search on jobs
        'job terminate': u'/jobs/{uuid}/terminate',  # POST -> terminate a job
        'job tasks': u'/jobs/{uuid}/tasks',  # GET -> tasks in job
        'tasks': u'/tasks',  # POST -> submit task
        'paginate tasks': u'/tasks/paginate',  # GET -> paginate tasks
        'paginate tasks summaries': u'/tasks/summaries/paginate',  # GET -> paginate tasks summaries;
        'tasks search': u'/tasks/search',  # POST -> make a custom search on tasks
        'task force': u'/tasks/force',  # POST -> force add
        'task update': u'/tasks/{uuid}',  # GET->result; DELETE -> abort, PATCH -> update resources
        'task snapshot': u'/tasks/{uuid}/snapshot/periodic',  # POST -> snapshots
        'task instant': u'/tasks/{uuid}/snapshot',  # POST -> get a snapshot
        'task stdout': u'/tasks/{uuid}/stdout',  # GET -> task stdout
        'task stderr': u'/tasks/{uuid}/stderr',  # GET -> task stderr
        'task abort': u'/tasks/{uuid}/abort',  # GET -> task
        'pools': u'/pools',  # POST -> submit pool
        'paginate pools': u'/pools/paginate',  # GET -> paginate pools
        'paginate pools summaries': u'/pools/summaries/paginate',  # GET -> paginate pools summaries
        'pools search': u'/pools/search',  # POST -> make a custom search on pools
        'pool close': u'/pools/{uuid}/close',  # POST -> close pool
        'pool update': u'/pools/{uuid}',  # GET -> pool, DELETE -> close & delete, PATCH -> update resources
        'user': u'/info',  # GET -> user info
        'profiles': u'/profiles',  # GET -> profiles list
        'profile details': u'/profiles/{profile}',  # GET -> profile details
        'settings': u'/settings',  # GET -> instance settings
    }
    return urls[key].format(**kwargs)


from ._version import get_versions  # noqa
__version__ = get_versions()['version']  # type: ignore
del get_versions

from .connection import Connection  # noqa



from typing import Any, Dict
import abc

# ********************************************************
# *******************  Filtering  ************************
# ********************************************************


class AbstractFiltering(metaclass=abc.ABCMeta):
    name: str = 'abstractFiltering'
    """
    Abstract base class for resources filtering, allowing to select only a subset of a resources bucket as task resources.
    """
    @abc.abstractmethod
    def to_json(self):
        """Get a dict ready to be json packed.

        :raises NotImplementedError: this is an abstract method, it should be overridden in child classes
        """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json):
        """Static method called to create the class obj from a json object.

        :param json: the json elements of the class
        :type json: Dict
        :raises NotImplementedError: this is an abstaract method, override it in it's child classes.
        """


class BucketPrefixFiltering(AbstractFiltering):
    name: str = 'prefixFiltering'

    """
        Allows to filter a resources bucket by name prefix. Only bucket files starting with the given prefix will be used as task resources.
    """
    def __init__(self, prefix: str):
        self.prefix = prefix

    @classmethod
    def from_json(cls, json: Dict[str, str]):
        """Static method called to create the class obj from a json object.

        :param json: the json elements of the class
        :type json: Dict
        "bucket_or_bucket_name""
        """
        prefix: str = json.get("prefix")
        return BucketPrefixFiltering(prefix)

    def __repr__(self) -> str:
        return "Filtering: {}  prefix: {}".format(self.name, self.prefix)

    def to_json(self) -> object:
        """Get a dict ready to be json packed.
        :return: the json elements of the class.
        :rtype: Dict

        """
        return {
            "prefix": self.prefix
        }


class Filtering(object):
    """
        Groups the various object filters on an advanced resources bucket.
    """

    def __init__(self):
        self._filters = {}

    def __repr__(self) -> str:
        return "[" + ",".join(map(str, self._filters.values())) + "]"

    def append(self, filtering: AbstractFiltering):
        """Add a new filtering object

        :param filtering: a filtering object of the bucket
        :type filtering: FilteringObject
        """
        self._filters[filtering.name] = filtering

    @classmethod
    def from_json(cls, json) -> 'Filtering':
        """Create the class sub objects of a Filtering from a json.

        :param json: the json elements of the class
        :type json: Dict
        """
        filtering = Filtering()
        for key in json:
            if BucketPrefixFiltering.name == key:
                filtering.append(BucketPrefixFiltering.from_json(json[key]))
        return filtering

    def to_json(self) -> Dict[str, Any]:
        """Get a dict ready to be json packed.
        """
        json = {}
        for key in self._filters:
            json[key] = self._filters[key].to_json()
        return json

# ********************************************************
# *************  ResourcesTransformation  ****************
# ********************************************************


class AbstractResourcesTransformation(metaclass=abc.ABCMeta):
    """
    Abstract base class for resources transformation, allowing to transform bucket objects before they are presented to the task as resources.
    """
    name: str = 'abstractResourcesTransformation'

    @abc.abstractmethod
    def to_json(self):
        """Get a dict ready to be json packed.
        """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json):
        """Static method called to create the class obj from a json object.

        :param json: the json elements of the class
        :type json: Dict
        """


class PrefixResourcesTransformation(AbstractResourcesTransformation):
    """
        Allows to remove a prefix from bucket files before they are presented to the task. During execution, the task will see files with paths stripped of the given prefix in its working directory.
    """
    name: str = 'stripPrefix'

    def __init__(self, prefix: str):
        """The PrefixResourcesTransformation constructor

        :param prefix: the prefix path of the resource bucket to be removed.
        :type prefix: str
        """
        self.prefix = prefix

    def __repr__(self) -> str:
        return "ResourcesTransformation: {}  prefix: {}".format(self.name, self.prefix)

    @classmethod
    def from_json(cls, json):
        """Create a new instance of the class using the given json object.

        :param json: The Json object representation.
        :type json: Dict
        :return: The PrefixResourcesTransformation new object
        :rtype: PrefixResourcesTransformation
        """
        return PrefixResourcesTransformation(json["prefix"])

    def to_json(self) -> object:
        """Get a dict ready to be json packed.
        """
        return {
            "prefix": self.prefix
        }


class ResourcesTransformation(object):
    """
        Groups the various object transformation on an advanced resources bucket.
    """
    def __init__(self):
        self._resource_transformers = {}

    def append(self, resource: AbstractResourcesTransformation):
        """Add a new resource transformation object

        :param filtering: a filtering object of the bucket
        :type filtering: FilteringObject
        """
        self._resource_transformers[resource.name] = resource

    def __repr__(self) -> str:
        return "[" + ",".join(map(str, self._resource_transformers.values())) + "]"

    @classmethod
    def from_json(cls, json) -> 'ResourcesTransformation':
        """Create the class sub objects of a ResourcesTransformation from a json.

        :param json: the json elements of the class
        :type json: Dict
        """
        resource = ResourcesTransformation()
        for key in json.keys():
            if PrefixResourcesTransformation.name == key:
                resource.append(
                    PrefixResourcesTransformation.from_json(json[key]))
        return resource

    def to_json(self) -> object:
        """Get a dict ready to be json packed.
        """
        json = {}
        for key in self._resource_transformers.keys():
            json[key] = self._resource_transformers[key].to_json()
        return json



from typing import List, Any


class PaginateResponse:
    """A paginate response

    :param is_truncated: Is the object truncated
    :type is_truncated: bool
    :param token: the token to used to retrieve the page, defaults to None
    :type token: str
    :param next_token: the next page token to be used to continue the pagination, defaults to None
    :type next_token: str
    :param page_data: the data objects retrieved
    :type page_data: List[Any]
    """

    def __init__(self, token: str, next_token: str, is_truncated: bool, page_data: List[Any]):
        self.token = token
        self.next_token = next_token
        self.is_truncated = is_truncated
        self.page_data = page_data



#!/usr/bin/env python

import qarnot
import pytest
from unittest.mock import patch, Mock, PropertyMock
import requests
import simplejson

expected_or_tags_filter = {"operator": "Or", "filters":[{"operator": "Equal", "field": "Tags", "value": "tag1"}, {"operator": "Equal", "field": "Tags", "value": "tag2"}]}
expected_and_tags_filter = {"operator": "And", "filters":[{"operator": "Equal", "field": "Tags", "value": "tag_inter1"}, {"operator": "Equal", "field": "Tags", "value": "tag_inter2"}]}
expected_and_or_tags_filter = {"operator": "And", "filters": [
            {"operator": "Or", "filters": [{"operator": "Equal", "field": "Tags", "value": "tag1"}, {
                "operator": "Equal", "field": "Tags", "value": "tag2"}]},
            {"operator": "And", "filters": [{"operator": "Equal", "field": "Tags", "value": "tag_inter1"}, {
                "operator": "Equal", "field": "Tags", "value": "tag_inter2"}]}
        ]}

class TestConnectionMethods:
    @pytest.mark.slow
    def test_connection_with_bad_ssl_return_the_good_exception(self):
        with pytest.raises(requests.exceptions.SSLError):
            assert qarnot.Connection(cluster_url="https://expired.badssl.com", client_token="token")

    @pytest.mark.slow
    def test_connection_with_bad_ssl_and_uncheck_return_JSONDecodeError_exception(self):
        with pytest.raises(simplejson.errors.JSONDecodeError):
            assert qarnot.Connection(cluster_url="https://expired.badssl.com", client_token="token", cluster_unsafe=True)

class TestConnectionPaginateMethods:
    @patch("qarnot.connection.Connection._get")
    def get_connection(self, mock_get):
        mock_get.return_value.status_code = 200
        connec = qarnot.Connection(
            client_token="token", cluster_url="https://localhost", storage_url="https://localhost")
        return connec

    def test_profiles_names(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._get") as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.json.return_value = ["test1", "test2"]
            retriever = connec.profiles_names()
            assert "/profiles" == mock_get.call_args[0][0]
            assert retriever[0] == "test1" and retriever[1] == "test2"

    def test_profiles(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.profiles_names") as mock_names:
            with patch("qarnot.connection.Connection._get") as mock_get:
                mock_names.return_value = ["hello01", "hello02", "hello03", "hello04", "hello05", "hello06", "hello07", "hello08", "hello09", "hello10", "hello11", "hello12", "hello13"]
                mock_get.return_value.status_code = 200
                mock_get.return_value.json.return_value = {"name":"world",
                                                           "constants": [{"name": "foo", "value": "bar"}, {"name": "foo2", "value": "bar2"}]}
                retriever = connec.profiles()
                assert len(retriever) == 13
                for ret in retriever:
                    assert ret.name == "world" and ret.constants == (('foo', 'bar'),('foo2', 'bar2'))
                for arg in mock_get.call_args_list:
                    assert arg[0][0].startswith("/profiles/hello")

    def test_profiles_with_internal_server_error(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.profiles_names") as mock_names:
            with patch("qarnot.connection.Connection._get") as mock_get:
                mock_names.return_value = ["hello01", "hello02", "hello03", "hello04", "hello05", "hello06", "hello07", "hello08", "hello09", "hello10", "hello11", "hello12", "hello13"]
                mock_get.return_value.status_code = 503
                mock_get.return_value.json.return_value = {"name":"world",
                                                           "constants": [{"name": "foo", "value": "bar"}, {"name": "foo2", "value": "bar2"}]}
                with pytest.raises(qarnot.QarnotGenericException):
                    retriever = connec.profiles()

    def test_profiles_with_not_found_error(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.profiles_names") as mock_names:
            with patch("qarnot.connection.Connection._get") as mock_get:
                mock_names.return_value = ["hello01", "hello02", "hello03", "hello04", "hello05", "hello06", "hello07", "hello08", "hello09", "hello10", "hello11", "hello12", "hello13"]
                mock_get.return_value.status_code = 404
                mock_get.return_value.json.return_value = {"name":"world",
                                                           "constants": [{"name": "foo", "value": "bar"}, {"name": "foo2", "value": "bar2"}]}
                retriever = connec.profiles()
                assert mock_get.call_args[0][0].startswith("/profiles/hello")
                assert len(retriever) == 0

    def return_mock_status(self, *args, **kwargs):
        errors = [1, 4, 6, 7, 12]
        status = 404 if self.test_profiles_with_5_error_step in errors else 200
        self.test_profiles_with_5_error_step += 1
        mock = Mock()
        type(mock).status_code = PropertyMock(return_value=status)
        mock.json.return_value = {"name": "world",
                                    "constants": [{"name": "foo", "value": "bar"}, {"name": "foo2", "value": "bar2"}]}
        return mock

    def test_profiles_with_5_not_found_error(self):
        self.test_profiles_with_5_error_step = 0
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.profiles_names") as mock_names:
            with patch("qarnot.connection.Connection._get") as mock_get:
                mock_names.return_value = ["hello01", "hello02", "hello03", "hello04", "hello05", "hello06", "hello07", "hello08", "hello09", "hello10", "hello11", "hello12", "hello13"]
                mock_get.side_effect = self.return_mock_status
                retriever = connec.profiles()
                assert mock_get.call_args[0][0].startswith("/profiles/hello")
                assert len(retriever) == 8

    def test_profile_details(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._get") as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.json.return_value = {"name":"world",
                                               "constants": [{"name": "foo", "value": "bar"}, {"name": "foo2", "value": "bar2"}]}
            retriever = connec.profile_details("hello")
            assert retriever.name == "world" and retriever.constants == (('foo', 'bar'),('foo2', 'bar2'))
            assert "/profiles/hello" == mock_get.call_args[0][0]

    def test_paginate_task_retriever_url(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}
            retriever = connec.tasks_page(summary=False)
            assert "/tasks/paginate" == mock_page_call.call_args[0][0]
            retriever = connec.tasks_page(summary=True)
            assert "/tasks/summaries/paginate" == mock_page_call.call_args[0][0]

    def test_paginate_job_retriever_url(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}
            retriever = connec.jobs_page()
            assert "/jobs/paginate" == mock_page_call.call_args[0][0]

    def test_paginate_pool_retriever_url(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}
            retriever = connec.pools_page(summary=False)
            assert "/pools/paginate" == mock_page_call.call_args[0][0]
            retriever = connec.pools_page(summary=True)
            assert "/pools/summaries/paginate" == mock_page_call.call_args[0][0]

    @pytest.mark.parametrize("tags, tags_intersect, expected_filter", [
        (["tag1", "tag2"], None, expected_or_tags_filter),
        (None, ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
        (["tag1", "tag2"], ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
    ])
    def test_paginate_task_retriever_filter(self, tags, tags_intersect, expected_filter):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}

            retriever = connec.tasks_page(tags=tags, tags_intersect=tags_intersect)
            assert expected_filter == mock_page_call.call_args[0][1]["filter"]

    @pytest.mark.parametrize("tags, tags_intersect, expected_filter", [
        (["tag1", "tag2"], None, expected_or_tags_filter),
        (None, ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
        (["tag1", "tag2"], ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
    ])
    def test_paginate_job_retriever_filter(self, tags, tags_intersect, expected_filter):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}

            retriever = connec.jobs_page(tags=tags, tags_intersect=tags_intersect)
            assert expected_filter == mock_page_call.call_args[0][1]["filter"]

    @pytest.mark.parametrize("tags, tags_intersect, expected_filter", [
        (["tag1", "tag2"], None, expected_or_tags_filter),
        (None, ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
        (["tag1", "tag2"], ["tag_inter1", "tag_inter2"], expected_and_tags_filter),
    ])
    def test_paginate_pool_retriever_filter(self, tags, tags_intersect, expected_filter):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}
            retriever = connec.pools_page(tags=tags, tags_intersect=tags_intersect)
            print(mock_page_call.call_args[0][1]["filter"])
            print(mock_page_call.call_args[0])
            print(mock_page_call.call_args)
            assert expected_filter == mock_page_call.call_args[0][1]["filter"]

    @pytest.mark.parametrize("token, max, expected_token, expected_max", [
        ("token", 1, "token", 1),
        (None, None, None, None),
    ])
    def test_paginate_task_retriever_paginate_page_values(self, token, max, expected_token, expected_max):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}

            retriever = connec.tasks_page(token=token, maximum=max)
            assert expected_token == mock_page_call.call_args[0][1]["token"]
            assert expected_max == mock_page_call.call_args[0][1]["maximumResults"]

    @pytest.mark.parametrize("token, max, expected_token, expected_max", [
        ("token", 1, "token", 1),
        (None, None, None, None),
    ])
    def test_paginate_job_retriever_paginate_request_values(self, token, max, expected_token, expected_max):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}

            retriever = connec.jobs_page(token=token, maximum=max)
            assert expected_token == mock_page_call.call_args[0][1]["token"]
            assert expected_max == mock_page_call.call_args[0][1]["maximumResults"]

    @pytest.mark.parametrize("token, max, expected_token, expected_max", [
        ("token", 1, "token", 1),
        (None, None, None, None),
    ])
    def test_paginate_pool_retriever_paginate_request_values(self, token, max, expected_token, expected_max):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":[]}
            retriever = connec.pools_page(token=token, maximum=max)
            assert expected_token == mock_page_call.call_args[0][1]["token"]
            assert expected_max == mock_page_call.call_args[0][1]["maximumResults"]

    def test_page_task_return_object(self):
        task_body = [{
            "uuid": "f78fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "default_name",
            "profile": "docker-bash",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "state": "submetted",
        }, {
            "uuid": "078fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "Second_default_name",
            "profile": "docker-bash2",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "state": "submetted",
        }]

        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":task_body}
            page_result = connec.tasks_page()
            assert page_result.token == "token"
            assert page_result.next_token == "nextToken"
            assert page_result.is_truncated == True
            result = page_result.page_data
            assert type(result[0]) == qarnot.task.Task
            assert result[0].uuid == "f78fdff8-7081-46e1-bb2f-d9cd4e185ece"
            assert result[1].uuid == "078fdff8-7081-46e1-bb2f-d9cd4e185ece"

    def test_page_job_return_object(self):
        job_body = [{
            "uuid": "f78fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "shortname": "f78fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "default_name",
            "profile": "docker-bash",
            "instanceCount": 1,
            "poolUuid": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "lastModified": "2019-11-08T10:54:11Z",
            "maxWallTime": "10:54:11",
            "state": "submetted",
            "useDependencies": False,
            "autoDeleteOnCompletion": False,
            "completionTimeToLive": None,
        }, {
            "uuid": "078fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "shortname": "078fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "Second_default_name",
            "profile": "docker-bash2",
            "instanceCount": 1,
            "poolUuid": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "lastModified": "2019-11-08T10:54:11Z",
            "maxWallTime": "10:54:11",
            "state": "submetted",
            "useDependencies": False,
            "autoDeleteOnCompletion": False,
            "completionTimeToLive": None,
        }]

        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token": "token", "nextToken": "nextToken", "isTruncated": True, "data": job_body}
            page_result = connec.jobs_page()
            assert page_result.token == "token"
            assert page_result.next_token == "nextToken"
            assert page_result.is_truncated == True
            result = page_result.page_data
            assert type(result[0]) == qarnot.job.Job
            assert result[0].uuid == "f78fdff8-7081-46e1-bb2f-d9cd4e185ece"
            assert result[1].uuid == "078fdff8-7081-46e1-bb2f-d9cd4e185ece"

    def test_page_pool_return_object(self):
        pool_body = [{
            "uuid": "f78fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "default_name",
            "profile": "docker-bash",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "state": "submetted",
        }, {
            "uuid": "078fdff8-7081-46e1-bb2f-d9cd4e185ece",
            "name": "Second_default_name",
            "profile": "docker-bash2",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "creationDate": "2019-11-08T10:54:11Z",
            "state": "submetted",
        }]

        connec = self.get_connection()
        with patch("qarnot.connection.Connection._page_call") as mock_page_call:
            mock_page_call.return_value = {"token" : "token","nextToken" : "nextToken","isTruncated":True,"data":pool_body}
            page_result = connec.pools_page()
            assert page_result.token == "token"
            assert page_result.next_token == "nextToken"
            assert page_result.is_truncated == True
            result = page_result.page_data
            assert type(result[0]) == qarnot.pool.Pool
            assert result[0].uuid == "f78fdff8-7081-46e1-bb2f-d9cd4e185ece"
            assert result[1].uuid == "078fdff8-7081-46e1-bb2f-d9cd4e185ece"

    def test_page_request(self):
        connec = self.get_connection()
        assert {"filter": None, "token": None, "maximumResults": None} == connec._paginate_request(None, None, None)
        assert {"filter": {}, "token": "token", "maximumResults": 12} == connec._paginate_request({}, "token", 12)

    def test_page_call_call_post(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._post") as mock_post:
            mock_post.return_value.json.return_value = {"token": "token", "nextToken": "nextToken", "isTruncated": True, "data": []}
            mock_post.return_value.status_code = 200
            connec._page_call("url", {})
            mock_post.assert_called_once_with("url", {})

    def test_page_call_raise_error_when_wrong_status_code(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._post") as mock_post:
            mock_post.return_value.status_code = 404
            mock_post.json.return_value = {"token": "token", "nextToken": "nextToken", "isTruncated": True, "data": []}
            with pytest.raises(qarnot.exceptions.QarnotGenericException):
                connec._page_call("url", {})

    def test_page_call_return_json(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._post") as mock_post:
            mock_post.return_value.status_code = 200
            json_ret = {"token": "token", "nextToken": "nextToken", "isTruncated": True, "data": []}
            mock_post.return_value.json.return_value = json_ret
            ret = connec._page_call("url", {})
            assert ret == json_ret

    def test_all_pages_break_when_false(self):
        connec = self.get_connection()
        mock1 = Mock()
        mock2 = Mock()
        mock3 = Mock()
        mock4 = Mock()
        mock5 = Mock()
        ret1 = Mock(next_token="next", is_truncated=True,page_data=[mock1, mock2])
        ret2 = Mock(next_token=None, is_truncated=False, page_data=[mock3, mock4])
        ret3 = Mock(next_token=None, is_truncated=False, page_data=[mock5])
        mock = Mock(side_effect=[ret1, ret2, ret3])
        ret = connec._all_pages(mock)
        assert mock1 in ret
        assert mock2 in ret
        assert mock3 in ret
        assert mock4 in ret
        assert mock5 not in ret

    def test_jobs(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.jobs_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            connec.jobs()
            mock_page_call.assert_called_once()

    def test_tasks(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.tasks_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            connec.tasks()
            mock_page_call.assert_called_once()

    def test_pools(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.pools_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            connec.pools()
            mock_page_call.assert_called_once()

    def test_all_jobs(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.jobs_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            iterator = connec.all_jobs()
            with pytest.raises(StopIteration):
                next(iterator)
            mock_page_call.assert_called_once()

    def test_all_tasks(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.tasks_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            iterator = connec.all_tasks()
            with pytest.raises(StopIteration):
                next(iterator)
            mock_page_call.assert_called_once()

    def test_all_pools(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection.pools_page") as mock_page_call:
            mock_page_call.return_value = qarnot.paginate.PaginateResponse("token", None, False, [])
            iterator = connec.all_pools()
            with pytest.raises(StopIteration):
                next(iterator)
            mock_page_call.assert_called_once()

    def test_user_information(self):
        connec = self.get_connection()
        with patch("qarnot.connection.Connection._get") as get_user:
            user_json = {
                "email":"",
                "maxBucket":5,
                "bucketCount":6,
                "quotaBytesBucket":7,
                "usedQuotaBytesBucket":8,
                "taskCount":9,
                "maxTask":10,
                "runningTaskCount":11,
                "maxRunningTask":12,
                "maxInstances":13,
                "maxPool":14,
                "poolCount":15,
                "maxRunningPool":16,
                "runningPoolCount":17,
                "runningInstanceCount":18,
                "runningCoreCount":19,
            }
            get_user.return_value.status_code = 200
            get_user.return_value.json.return_value = user_json
            user = connec.user_info
            assert user.email == user_json.get('email', '')
            assert user.max_bucket == user_json['maxBucket']
            assert user.bucket_count == user_json.get('bucketCount', -1)
            assert user.quota_bytes_bucket == user_json['quotaBytesBucket']
            assert user.used_quota_bytes_bucket == user_json['usedQuotaBytesBucket']
            assert user.task_count == user_json['taskCount']
            assert user.max_task == user_json['maxTask']
            assert user.running_task_count == user_json['runningTaskCount']
            assert user.max_running_task == user_json['maxRunningTask']
            assert user.max_instances == user_json['maxInstances']
            assert user.max_pool == user_json['maxPool']
            assert user.pool_count == user_json['poolCount']
            assert user.max_running_pool == user_json['maxRunningPool']
            assert user.running_pool_count == user_json['runningPoolCount']
            assert user.running_instance_count == user_json['runningInstanceCount']
            assert user.running_core_count == user_json['runningCoreCount']



"""Module for bucket object."""

# Copyright 2017 Qarnot computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import hashlib
import os
import posixpath
import shutil
import itertools
import deprecation
from . import __version__
from typing import Optional

from boto3.s3.transfer import TransferConfig
from itertools import groupby
from operator import attrgetter

from . import _util
from .exceptions import BucketStorageUnavailableException
from .storage import Storage

from .advanced_bucket import Filtering, ResourcesTransformation
# Max size in bytes before uploading in parts.

AWS_UPLOAD_MAX_SIZE = 8 * 1024 * 1024
# Size of parts when uploading in parts
AWS_UPLOAD_PART_SIZE = 8 * 1024 * 1024

s3_multipart_config = TransferConfig(
    multipart_threshold=AWS_UPLOAD_MAX_SIZE,
    multipart_chunksize=AWS_UPLOAD_PART_SIZE,
    max_concurrency=10,
    num_download_attempts=10,
)


class Bucket(Storage):  # pylint: disable=W0223
    """Represents a resource/result bucket.

    This class is the interface to manage resources or results from a
    :class:`qarnot.bucket.Bucket`.

    .. note::
       A :class:`Bucket` must be created with
       :meth:`qarnot.connection.Connection.create_bucket`
       or retrieved with :meth:`qarnot.connection.Connection.buckets`, :meth:`qarnot.connection.Connection.retrieve_bucket`,
       or :meth:`qarnot.connection.Connection.retrieve_or_create_bucket`.

    .. note::
       Paths given as 'remote' arguments,
       (or as path arguments for :func:`Bucket.directory`)
       **must** be valid unix-like paths.
    """

    def __init__(self, connection, name, create=True, filtering: Filtering = None, resources_transformation: ResourcesTransformation = None):
        super().__init__()
        if connection.s3client is None:
            raise BucketStorageUnavailableException()

        self._connection = connection
        self._uuid = name
        self._filtering: Optional[Filtering] = filtering or Filtering()
        self._resources_transformation: Optional[ResourcesTransformation] = \
            resources_transformation or ResourcesTransformation()

        if create:
            self._connection.s3client.create_bucket(Bucket=name)

    def to_json(self):
        """Get a dict ready to be json packed from this bucket."""
        as_json_dict = dict()
        as_json_dict['bucketName'] = self._uuid
        as_json_dict['filtering'] = None if self._filtering is None else self._filtering.to_json()
        as_json_dict['resourcesTransformation'] = (None if self._resources_transformation is None
                                                   else self._resources_transformation.to_json())
        return as_json_dict

    @classmethod
    def from_json(cls, connection, json_bucket):
        """Create a Bucket object from a json advance bucket.
        :param qarnot.connection.Connection connection: the cluster connection
        :param dict json_bucket: Dictionary representing the bucket
        :returns: The created :class:`~qarnot.bucket.Bucket`.
        """
        filtering = Filtering.from_json(json_bucket['filtering'])
        resource_transformation = ResourcesTransformation.from_json(json_bucket['resourcesTransformation'])
        bucket = Bucket(connection, json_bucket['bucketName'], create=False, filtering=filtering, resources_transformation=resource_transformation)
        return bucket

    def with_filtering(self, filtering):
        """Create a new Bucket object from the given bucket with a specific filtering.
        examples :
        * new_bucket = bucket.with_filtering(BucketPrefixFiltering("prefix1"))
        * new_bucket = Bucket(connection, "name", False).with_filtering(BucketPrefixFiltering("prefix1"))

        :param :class:`~qarnot.advance_bucket.AbstractFiltering` filtering: Filtering to add to the bucket.
        :returns: The created :class:`~qarnot.bucket.Bucket`.
        """
        bucket_copy = Bucket(self._connection, self._uuid,
                             create=False, filtering=Filtering(), resources_transformation=self._resources_transformation)
        bucket_copy._filtering.append(filtering)
        return bucket_copy

    def with_resource_transformation(self, resource):
        """Create a new Bucket object from the given bucket with a specific resource transformation.
        examples :
        * new_bucket = Bucket(connection, "name", False).with_resource_transformation(PrefixResourcesTransformation("prefix2"))
        * new_bucket = bucket.with_resource_transformation(PrefixResourcesTransformation("prefix2")).with_filtering(BucketPrefixFiltering("prefix1"))

        :param :class:`~qarnot.advance_bucket.AbstractResourcesTransformation` resource: The resource transformation to add to the bucket.
        :returns: The created :class:`~qarnot.bucket.Bucket`.
        """
        bucket_copy = Bucket(self._connection, self._uuid,
                             create=False, filtering=self._filtering, resources_transformation=ResourcesTransformation())
        bucket_copy._resources_transformation.append(resource)
        return bucket_copy

    @classmethod
    def _retrieve(cls, connection, bucket_uuid):
        """Retrieve information of a bucket on a cluster.

        :param :class:`qarnot.connection.Connection` connection: the cluster
            to get the bucket from
        :param str bucket_uuid: the UUID of the bucket to retrieve

        :rtype: :class:`qarnot.bucket.Bucket`
        :returns: The retrieved bucket.

        :raises qarnot.exceptions.BucketStorageUnavailableException: the bucket storage engine is not available
        """
        return connection.retrieve_bucket(uuid=bucket_uuid)

    def delete(self):
        """ Delete the bucket represented by this :class:`Bucket`."""

        n = 1000  # delete object count max request

        bucket = self._connection.s3resource.Bucket(self._uuid)
        versioned_bucket = self._connection.s3resource.BucketVersioning(self._uuid)

        if versioned_bucket.status == 'None':
            objectlist = list(bucket.objects.all())
            listofobjectlist = [[{'Key': x.key} for x in objectlist[i:i + n]] for i in range(0, len(objectlist), n)]
        else:
            objectlist = list(bucket.object_versions.all())
            listofobjectlist = [[{'Key': x.key, 'VersionId': x.id} for x in objectlist[i:i + n]] for i in range(0, len(objectlist), n)]

        for item in listofobjectlist:
            bucket.delete_objects(
                Delete={
                    'Objects': item
                }
            )
        self._connection.s3client.delete_bucket(Bucket=self._uuid)

    def list_files(self):
        """List files in the bucket

        :rtype: list(:class:`S3.ObjectSummary`)
        :returns: A list of ObjectSummary resources

        """
        bucket = self._connection.s3resource.Bucket(self._uuid)
        return bucket.objects.all()

    def directory(self, directory=''):
        """List files in a directory of the bucket according to prefix.

        :rtype: list(:class:`S3.ObjectSummary`)
        :returns: A list of ObjectSummary resources
        """
        bucket = self._connection.s3resource.Bucket(self._uuid)
        return bucket.objects.filter(Prefix=directory)

    def sync_directory(self, directory, verbose=False, remote=None):
        """Synchronize a local directory with the remote buckets.

        :param str directory: The local directory to use for synchronization
        :param bool verbose: Print information about synchronization operations
        :param str remote: path of the directory on remote node (defaults to *local*)

        .. warning::
           Local changes are reflected on the server, a file present on the
           bucket but not in the local directory will be deleted from the bucket.

           A file present in the directory but not in the bucket will be uploaded.

        .. note::
           The following parameters are used to determine whether
           synchronization is required :

              * name
              * size
              * sha1sum
        """
        if not directory.endswith(os.sep):
            directory += os.sep

        filesdict = {}
        for root, _, files in os.walk(directory):
            root = _util.decode(root)
            files = list(map(_util.decode, files))

            for file_ in files:
                filepath = os.path.join(root, file_)
                name = filepath[len(directory):]
                name = name.replace(os.sep, '/')
                filesdict[name] = filepath

        self.sync_files(filesdict, verbose, remote)

    def sync_files(self, files, verbose=False, remote=None):
        """Synchronize files  with the remote buckets.

        :param dict files: Dictionary of synchronized files
        :param bool verbose: Print information about synchronization operations
        :param str remote: path of the directory on remote node (defaults to *local*)

        Dictionary key is the remote file path while value is the local file
        path.

        .. warning::
           Local changes are reflected on the server, a file present on the
           bucket but
           not in the local directory will be deleted from the bucket.

           A file present in the directory but not in the bucket will be uploaded.

        .. note::
           The following parameters are used to determine whether
           synchronization is required :

              * name
              * size
              * sha1sum
        """

        class Comparable(object):
            def __init__(self, name_, e_tag, filepath_):
                self.name = name_
                self.e_tag = e_tag
                self.filepath = filepath_

            def __repr__(self):
                return "Name {0}, ETag {1}".format(self.name, self.e_tag)

            def __eq__(self, other):
                return self.name == other.name and self.e_tag == other.e_tag

            def __hash__(self):
                return hash(self.name) ^ hash(self.e_tag)

        def aws_md5sum(sourcepath):
            if os.stat(sourcepath).st_size < AWS_UPLOAD_MAX_SIZE:
                hash_md5 = hashlib.md5()
                with open(sourcepath, "rb") as f:
                    for chunk in iter(lambda: f.read(4096), b""):
                        hash_md5.update(chunk)
                return "\"{0}\"".format(hash_md5.hexdigest())
            else:
                md5s = []
                with open(sourcepath, 'rb') as fp:
                    while True:

                        data = fp.read(AWS_UPLOAD_PART_SIZE)

                        if not data:
                            break
                        md5s.append(hashlib.md5(data))

                digests = b"".join(m.digest() for m in md5s)

                new_md5 = hashlib.md5(digests)
                return "\"{0}-{1}\"".format(new_md5.hexdigest(), len(md5s))

        def localtocomparable(name_, filepath_, remote):
            if remote is not None:
                name_ = os.path.join(remote, name_.lstrip('/'))
            return Comparable(name_.replace(os.sep, '/'), aws_md5sum(filepath_), filepath_)

        def objectsummarytocomparable(object_):
            return Comparable(object_.key, object_.e_tag, None)

        localfiles = set()
        for name, filepath in files.items():
            localfiles.add(localtocomparable(name.replace(os.path.sep, '/'), filepath, remote))

        remotefiles = set(map(objectsummarytocomparable, self.list_files()))

        adds = localfiles - remotefiles
        removes = remotefiles - localfiles

        for file_ in removes:
            if remote is not None and not file_.name.startswith(remote):
                continue
            renames = (x for x in adds if x.e_tag == file_.e_tag and all(rem.name != x.name for rem in remotefiles))
            for dup in renames:
                if verbose:
                    print("Copy", file_.name, "to", dup.name)
                self.copy_file(file_.name, dup.name)
            if verbose:
                print("Remove:", file_.name)
            self.delete_file(file_.name)

        remotefiles = set(map(objectsummarytocomparable, self.list_files()))

        sadds = sorted(adds, key=lambda x: x.e_tag)
        groupedadds = (list(g) for _, g in itertools.groupby(sadds, lambda x: x.e_tag))

        for entry in groupedadds:
            try:
                rem = next(x for x in remotefiles if x.e_tag == entry[0].e_tag)
                if rem.name == entry[0].name:
                    continue
                if verbose:
                    print("Copy", rem.name, "to", entry[0].name)
                self.copy_file(rem.name, entry[0].name)
            except StopIteration:
                if verbose:
                    print("Upload:", entry[0].filepath, '->', entry[0].name)
                self.add_file(entry[0].filepath, entry[0].name)
            for link in entry[1:]:  # duplicate files
                if verbose:
                    print("Copy", entry[0].name, "to", link.name)
                self.copy_file(entry[0].name, link.name)

    @_util.copy_docs(Storage.add_file)
    def add_file(self, local_or_file, remote=None):
        tobeclosed = False
        if _util.is_string(local_or_file):
            file_ = open(local_or_file, 'rb')
            tobeclosed = True
        else:
            file_ = local_or_file
        dest = remote or os.path.basename(file_.name)

        self._connection.s3client.upload_fileobj(file_, self._uuid, dest, Config=s3_multipart_config)
        if tobeclosed:
            file_.close()

    @_util.copy_docs(Storage.get_all_files)
    def get_all_files(self, output_dir, progress=None):
        list_files_only = [x for x in self.list_files() if not x.key.endswith('/')]
        list_directories_only = [x for x in self.list_files() if x.key.endswith('/')]

        for directory in list_directories_only:
            if not os.path.isdir(os.path.join(output_dir, directory.key.lstrip('/'))):
                os.makedirs(os.path.join(output_dir, directory.key.lstrip('/')))

        for _, dupes in groupby(sorted(list_files_only, key=attrgetter('e_tag')), attrgetter('e_tag')):
            file_info = next(dupes)
            first_file = os.path.join(output_dir, file_info.key.lstrip('/'))
            self.get_file(file_info.get()['Body'], local=first_file)  # avoids making a useless HEAD request

            for dupe in dupes:
                local = os.path.join(output_dir, dupe.key.lstrip('/'))
                directory = os.path.dirname(local)
                if not os.path.exists(directory):
                    os.makedirs(directory)
                if (os.path.abspath(os.path.realpath(local)) is not os.path.abspath(os.path.realpath(first_file))):
                    shutil.copy(first_file, local)

    @_util.copy_docs(Storage.get_file)
    def get_file(self, remote, local=None, progress=None):
        return super(Bucket, self).get_file(remote, local, progress)

    @_util.copy_docs(Storage.add_directory)
    def add_directory(self, local, remote=""):
        if not os.path.isdir(local):
            raise IOError("Not a valid directory")
        if remote and not remote.endswith('/'):
            remote += '/'
        for dirpath, _, files in os.walk(local):
            dirpath = _util.decode(dirpath)
            files = list(map(_util.decode, files))

            remote_loc = dirpath.replace(local, remote, 1)
            for filename in files:
                self.add_file(os.path.join(dirpath, filename),
                              posixpath.join(remote_loc, filename))

    @_util.copy_docs(Storage.copy_file)
    def copy_file(self, source, dest):
        copy_source = {
            'Bucket': self._uuid,
            'Key': source
        }
        return self._connection.s3client.copy_object(CopySource=copy_source, Bucket=self._uuid, Key=dest)

    @deprecation.deprecated(deprecated_in="2.6.0", removed_in="3.0",
                            current_version=__version__,  # type: ignore
                            details="Legacy function")
    @_util.copy_docs(Storage.flush)
    def flush(self):
        pass

    @deprecation.deprecated(deprecated_in="2.6.0", removed_in="3.0",
                            current_version=__version__,  # type: ignore
                            details="Legacy function")
    @_util.copy_docs(Storage.update)
    def update(self, flush=False):
        pass

    def _download_file(self, remote, local, progress=None):
        with open(local, 'wb') as data:
            if hasattr(remote, 'read'):
                shutil.copyfileobj(remote, data)
            else:
                self._connection.s3client.download_fileobj(self._uuid, remote, data)
        return local

    @_util.copy_docs(Storage.delete_file)
    def delete_file(self, remote):
        self._connection.s3client.delete_object(Bucket=self._uuid, Key=remote)

    @property
    def uuid(self):
        """ Bucket identifier"""
        return self._uuid

    @property
    def description(self):
        """ Bucket identifier"""
        return self._uuid



import datetime
import pytest

from qarnot.task import Task
from qarnot.bucket import Bucket
from qarnot.advanced_bucket import BucketPrefixFiltering, PrefixResourcesTransformation
import datetime

from .mock_connection import MockConnection
from .mock_task import default_json_task, task_with_running_instances

@pytest.fixture(name="mock_conn")
def mock_conn_fixture():
    return MockConnection()


class TestTaskProperties:
    def submit_task(self, task):
        task._uuid = "set"

    def test_task_autodelete_default_value(self, mock_conn):
        task = Task(mock_conn, "task-name")
        assert task.auto_delete is False

    def test_task_completion_ttl_default_value(self, mock_conn):
        task = Task(mock_conn, "task-name")
        assert "00:00:00" == task.completion_ttl

    def test_task_autodelete_set_get(self, mock_conn):
        task = Task(mock_conn, "task-name")
        task.auto_delete = False
        assert task.auto_delete is False
        task.auto_delete = True
        assert task.auto_delete is True

    def test_task_completion_ttl_set_get(self, mock_conn):
        task = Task(mock_conn, "task-name")
        task.completion_ttl = datetime.timedelta(days=2, hours=33, minutes=66, seconds=66)
        assert "3.10:07:06" == task.completion_ttl
        task.completion_ttl = "4.11:08:06"
        assert "4.11:08:06" == task.completion_ttl

    def test_task_are_in_task_to_json(self, mock_conn):
        task = Task(mock_conn, "task-name")
        task.completion_ttl = "4.11:08:06"
        task.auto_delete = True
        json_task = task._to_json()  # pylint: disable=protected-access

        assert json_task['completionTimeToLive'] == '4.11:08:06'
        assert json_task['autoDeleteOnCompletion'] is True

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", None),
        ("state_transition_time", None),
        ("previous_state_transition_time", None),
        ("last_modified", None),
        ("snapshot_interval", None),
        ("progress", None),
        ("execution_time", None),
        ("wall_time", None),
        ("end_date", None),
    ])
    def test_task_property_default_value(self, mock_conn, property_name,  expected_value):
        task = Task(mock_conn, "task-name")
        assert getattr(task, property_name) is expected_value

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", default_json_task["previousState"]),
        ("state_transition_time", default_json_task["stateTransitionTime"]),
        ("previous_state_transition_time", default_json_task["previousStateTransitionTime"]),
        ("last_modified", default_json_task["lastModified"]),
        ("snapshot_interval", default_json_task["snapshotInterval"]),
        ("progress", default_json_task["progress"]),
        ("execution_time", default_json_task["executionTime"]),
        ("wall_time", default_json_task["wallTime"]),
        ("end_date", default_json_task["endDate"]),
    ])
    def test_task_property_update_value(self, mock_conn, property_name,  expected_value):
        task = Task(mock_conn, "task-name")
        task._update(default_json_task)
        assert getattr(task, property_name) is expected_value

    @pytest.mark.parametrize("property_name, expected_value", [
        ("name", default_json_task["name"]),
    ])
    def test_task_property_send_to_json_representation(self, mock_conn, property_name, expected_value):
        task = Task(mock_conn, "task-name")
        task._update(default_json_task)
        task_json = task._to_json()
        assert task_json[property_name] is expected_value

    @pytest.mark.parametrize("property_name, set_value, expected_value", [
        ("name", "name", "name")
    ])
    def test_task_set_property_value(self, mock_conn, property_name, set_value, expected_value):
        task = Task(mock_conn, "task-name")
        setattr(task, property_name, set_value)
        assert getattr(task, property_name) == expected_value

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("uuid", "error-can-set-uuid", AttributeError)
    ])
    def test_task_set_forbidden_property_raise_exception(self, mock_conn, property_name, set_value, exception):
        task = Task(mock_conn, "task-name")
        with pytest.raises(exception):
            setattr(task, property_name, set_value)

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("instancecount", 5, AttributeError)
    ])
    def test_task_set_property_raise_exception_after_submitted(self, mock_conn, property_name, set_value, exception):
        task = Task(mock_conn, "task-name")
        self.submit_task(task)
        with pytest.raises(exception):
            setattr(task, property_name, set_value)

    def test_advance_bucket_in_task_to_json(self, mock_conn):
        task = Task(mock_conn, "task-name", "profile")
        bucket = Bucket(mock_conn, "name", False)
        bucket2 = bucket.with_filtering(BucketPrefixFiltering(
            "prefix1")).with_resource_transformation(PrefixResourcesTransformation("prefix2"))

        task.resources.append(bucket2)
        json_task = task._to_json()
        json_bucket = json_task["advancedResourceBuckets"][0]
        assert "name" == json_bucket["bucketName"]
        assert "prefix1" == json_bucket["filtering"]["prefixFiltering"]["prefix"]
        assert "prefix2" == json_bucket["resourcesTransformation"]["stripPrefix"]["prefix"]

    def test_bucket_in_task_from_json(self, mock_conn):
        json_bucket = "bucket-name"
        json_task = {
            "name": "taskName",
            "shortname": "taskShortname",
            "profile": "profile",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "resourceBuckets": [json_bucket],
            "creationDate": "2019-11-08T10:54:11Z",
            "uuid": "000",
            "state": "Submitted",
        }
        task = Task.from_json(mock_conn, json_task)
        task._auto_update = False
        assert "bucket-name" == task.resources[0].uuid

    def test_advance_bucket_in_task_from_json(self, mock_conn):
        json_bucket = {
            "bucketName": "name",
            "filtering": {
                "prefixFiltering": {
                    "prefix": "prefix1"
                }
            },
            "resourcesTransformation": {
                "stripPrefix": {
                    "prefix": "prefix2"
                }
            }
        }
        json_task = {
            "name": "taskName",
            "shortname": "taskShortname",
            "profile": "profile",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "advancedResourceBuckets": [json_bucket],
            "creationDate": "2019-11-08T10:54:11Z",
            "uuid": "000",
            "state": "Submitted",
        }
        task = Task.from_json(mock_conn, json_task)
        task._auto_update = False

        assert "name" == task.resources[0].uuid
        assert "prefix1" == task.resources[0]._filtering._filters["prefixFiltering"].prefix
        assert "prefix2" == task.resources[0]._resources_transformation._resource_transformers["stripPrefix"].prefix


    def test_execution_attempt_count_in_running_instances(self, mock_conn):
        task = Task(mock_conn, "task-name")
        task._update(task_with_running_instances)
        assert len(task.status.running_instances_info.per_running_instance_info) == 2
        assert task.status.running_instances_info.per_running_instance_info[0].execution_attempt_count == 1
        assert task.status.running_instances_info.per_running_instance_info[1].execution_attempt_count == 2


    def test_execution_attempt_count_in_completed_instances(self, mock_conn):
        task = Task(mock_conn, "task-name")
        task._update(default_json_task)
        assert task.completed_instances[0].execution_attempt_count == 43



import pytest
import qarnot
from qarnot.pool import Pool
from qarnot.bucket import Bucket
from qarnot.advanced_bucket import BucketPrefixFiltering, PrefixResourcesTransformation
import datetime
from .mock_connection import MockConnection, PatchRequest, none_function
from .mock_pool import default_json_pool


class TestPoolProperties:
    def submit_pool(self, pool):
        pool._uuid = "set"

    conn = MockConnection()
    def test_pool_autodelete_default_value(self):
        pool = Pool(self.conn, "pool-name", "profile")
        assert False == pool.auto_delete

    def test_pool_completion_ttl_default_value(self):
        pool = Pool(self.conn, "pool-name", "profile")
        assert "00:00:00" == pool.completion_ttl

    def test_pool_autodelete_set_get(self):
        pool = Pool(self.conn, "pool-name", "profile")
        pool.auto_delete = False
        assert False == pool.auto_delete
        pool.auto_delete = True
        assert True == pool.auto_delete

    def test_pool_completion_ttl_set_get(self):
        pool = Pool(self.conn, "pool-name", "profile")
        pool.completion_ttl = datetime.timedelta(days=2, hours=33, minutes=66, seconds=66)
        assert "3.10:07:06" == pool.completion_ttl
        pool.completion_ttl = "4.11:08:06"
        assert "4.11:08:06" == pool.completion_ttl

    def test_pool_are_in_pool_to_json(self):
        pool = Pool(self.conn, "pool-name", "profile")
        pool.completion_ttl = "4.11:08:06"
        pool.auto_delete = True
        json_pool = pool._to_json()

        assert json_pool['completionTimeToLive'] == '4.11:08:06'
        assert json_pool['autoDeleteOnCompletion'] == True

    def test_update_resources_send_the_good_url(self):
        update_connection = MockConnection()
        pool = Pool(update_connection, "pool-name", "profile")
        pool._uuid = "uuid"
        pool.update = none_function
        pool.update_resources()
        assert type(update_connection.requests[0]) == PatchRequest
        assert update_connection.requests[0].uri == "/pools/uuid"

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", None),
        ("state_transition_time", None),
        ("previous_state_transition_time", None),
        ("last_modified", None),
        ("execution_time", None),
        ("end_date", None),
        ("tasks_default_wait_for_pool_resources_synchronization", False),
    ])
    def test_pool_property_default_value(self, property_name,  expected_value):
        pool = Pool(self.conn, "pool-name", "profile")
        assert getattr(pool, property_name) is expected_value

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", default_json_pool["previousState"]),
        ("state_transition_time", default_json_pool["stateTransitionTime"]),
        ("previous_state_transition_time", default_json_pool["previousStateTransitionTime"]),
        ("last_modified", default_json_pool["lastModified"]),
        ("execution_time", default_json_pool["executionTime"]),
        ("end_date", default_json_pool["endDate"]),
        ("tasks_default_wait_for_pool_resources_synchronization", default_json_pool["taskDefaultWaitForPoolResourcesSynchronization"]),
    ])
    def test_pool_property_update_value(self, property_name,  expected_value):
        pool = Pool(self.conn, "pool-name", "profile")
        pool._update(default_json_pool)
        assert getattr(pool, property_name) is expected_value

    @pytest.mark.parametrize("property_name, expected_value", [
        ("taskDefaultWaitForPoolResourcesSynchronization", default_json_pool["taskDefaultWaitForPoolResourcesSynchronization"]),
    ])
    def test_pool_property_send_to_json_representation(self, property_name, expected_value):
        pool = Pool(self.conn, "pool-name", "profile")
        pool._update(default_json_pool)
        pool_json = pool._to_json()
        assert pool_json[property_name] is expected_value

    @pytest.mark.parametrize("property_name, set_value, expected_value", [
        ("name", "name", "name")
    ])
    def test_pool_set_property_value(self, property_name, set_value, expected_value):
        pool = Pool(self.conn, "pool-name", "profile")
        setattr(pool, property_name, set_value)
        assert getattr(pool, property_name) == expected_value

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("uuid", "error-can-set-uuid", AttributeError)
    ])
    def test_pool_set_forbidden_property_raise_exception(self, property_name, set_value, exception):
        pool = Pool(self.conn, "pool-name", "profile")
        with pytest.raises(exception):
            setattr(pool, property_name, set_value)

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("instancecount", 5, AttributeError)
    ])
    def test_pool_set_property_raise_exception_after_submitted(self, property_name, set_value, exception):
        pool = Pool(self.conn, "pool-name", "profile")
        self.submit_pool(pool)
        with pytest.raises(exception):
            setattr(pool, property_name, set_value)

    def test_advance_bucket_in_pool_to_json(self):
        pool = Pool(self.conn, "pool-name", "profile")
        bucket = Bucket(self.conn, "name", False)
        bucket2 = bucket.with_filtering(BucketPrefixFiltering(
            "prefix1")).with_resource_transformation(PrefixResourcesTransformation("prefix2"))

        pool.resources.append(bucket2)
        json_pool = pool._to_json()
        json_bucket = json_pool["advancedResourceBuckets"][0]
        assert "name" == json_bucket["bucketName"]
        assert "prefix1" == json_bucket["filtering"]["prefixFiltering"]["prefix"]
        assert "prefix2" == json_bucket["resourcesTransformation"]["stripPrefix"]["prefix"]

    def test_bucket_in_pool_from_json(self):
        json_bucket = "bucket-name"
        json_pool = {
            "name": "poolName",
            "shortname": "poolShortname",
            "profile": "profile",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "resourceBuckets": [json_bucket],
            "creationDate": "2019-11-08T10:54:11Z",
            "uuid": "000",
            "state": "Submitted",
        }
        pool = Pool.from_json(self.conn, json_pool)
        pool._auto_update = False
        assert "bucket-name" == pool.resources[0].uuid

    def test_advance_bucket_in_pool_from_json(self):
        json_bucket = {
            "bucketName": "name",
            "filtering": {
                "prefixFiltering": {
                    "prefix": "prefix1"
                }
            },
            "resourcesTransformation": {
                "stripPrefix": {
                    "prefix": "prefix2"
                }
            }
        }
        json_pool = {
            "name":"poolName",
            "shortname":"poolShortname",
            "profile": "profile",
            "instanceCount": 1,
            "runningCoreCount": None,
            "runningInstanceCount": None,
            "advancedResourceBuckets": [json_bucket],
            "creationDate": "2019-11-08T10:54:11Z",
            "uuid": "000",
            "state": "Submitted",
        }
        pool = Pool.from_json(self.conn, json_pool)
        pool._auto_update = False

        assert "name" == pool.resources[0].uuid
        assert "prefix1" == pool.resources[0]._filtering._filters["prefixFiltering"].prefix
        assert "prefix2" == pool.resources[0]._resources_transformation._resource_transformers["stripPrefix"].prefix



"""Module to handle a job."""

# Copyright 2017 Qarnot computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import datetime


from .task import Task
from . import get_url, raise_on_error, _util
from .exceptions import MaxJobException, NotEnoughCreditsException, MissingJobException, UnauthorizedException


class JobState:
    Active = "Active"
    Terminating = "Terminating"
    Completed = "Completed"
    Deleting = "Deleting"


class Job(object):
    """Represents a Qarnot job.

    .. note::
       A :class:`Job` must be created with
       :meth:`qarnot.connection.Connection.create_job`
       or retrieved with :meth:`qarnot.connection.Connection.jobs` or :meth:`qarnot.connection.Connection.retrieve_job`.
    """
    def __init__(self, connection, name, pool=None, shortname=None, use_dependencies=False):
        """Create a new :class:`Job`.

        :param connection: the cluster on which to send the job
        :type connection: :class:`qarnot.connection.Connection`
        :param name: given name of the job
        :type name: :class:`str`
        :param pool: which Pool to submit the job in,
        :type pool: :class:`~qarnot.pool.Pool` or None
        :param shortname: userfriendly job name
        :type shortname: :class:`str`
        :param use_dependencies: allow dependencies between tasks in this job
        :type job: :class:`bool`
        """
        self._connection = connection
        self._name = name
        self._shortname = shortname
        self._pool_uuid = None
        if pool is not None:
            if _util.is_string(pool):
                self._pool_uuid = pool
            else:
                self._pool_uuid = pool.uuid
        self._state = ""
        self._uuid = None
        self._creation_date = datetime.datetime.now()
        self._use_dependencies = use_dependencies
        self._max_wall_time = None
        self._update_cache_time = 5
        self._auto_update = True
        self._last_auto_update_state = self._auto_update
        self._tags = []

        self._last_modified = None
        self._last_cache = time.time()
        self._completion_time_to_live = "00:00:00"
        self._auto_delete = False
        self._previous_state = None
        self._state_transition_time = None
        self._previous_state_transition_time = None

    @property
    def auto_update(self):
        """:type: :class:`bool`

        :getter: Returns this job's auto update state
        :setter: Sets this job's auto update state

        Auto update state, default to True
        When auto update is disabled properties will always return cached value
        for the object and a call to :meth:`update` will be required to get latest values from the REST Api.
        """
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value):
        """Setter for auto_update feature
        """
        self._auto_update = value
        self._last_auto_update_state = self._auto_update

    @property
    def update_cache_time(self):
        """:type: :class:`int`

        :getter: Returns this job's auto update state
        :setter: Sets this job's auto update state

        Cache expiration time, default to 5s
        """
        return self._update_cache_time

    @property
    def state(self):
        """:type: :class:`str`
        :getter: return this job's state

        State of the job.

        Value is in
           * UnSubmitted
           * Active,
           * Terminating,
           * Completed,
           * Deleting

        .. warning::
           this is the state of the job when the object was retrieved,
           call :meth:`update` for up to date value.
        """
        if self._auto_update:
            self.update()
        return self._state

    @property
    def tasks(self):
        """:type: List of :class:`~qarnot.task.Task`
        :getter: Returns this job tasks

        The tasks submitted in this job.
        """
        if self._uuid is None:
            return
        response = self._connection._get(get_url('job tasks', uuid=self._uuid))
        if response.status_code == 404:
            raise MissingJobException(response.json()['message'])
        raise_on_error(response)
        return [Task.from_json(self, task, True) for task in response.json()]

    @property
    def use_dependencies(self):
        """:type: :class:`bool`
        :getter: task's job can have dependencies
        :setter: Set if there is task's job dependencies

        Can be set until :meth:`submit` is called.
        """
        return self._use_dependencies

    @use_dependencies.setter
    def use_dependencies(self, value):
        """setter for job"""
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        else:
            self._use_dependencies = value

    @property
    def uuid(self):
        """:type: :class:`str`
        :getter: Returns this job's uuid

        The job's uuid.

        Automatically set when a job is submitted.
        """
        return self._uuid

    @property
    def name(self):
        """:type: :class:`str`
        :getter: Returns this job's name
        :setter: Sets this job's name

        The job's name.

        Can be set until job is submitted.
        """
        return self._name

    @name.setter
    def name(self, value):
        """Setter for name."""
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        else:
            self._name = value

    @property
    def shortname(self):
        """:type: :class:`str`
        :getter: Returns this job's shortname
        :setter: Sets this job's shortname

        The job's shortname, must be DNS compliant and unique, if not provided, will default to :attr:`uuid`.

        Can be set until job is submitted.
        """
        return self._shortname

    @shortname.setter
    def shortname(self, value):
        """Setter for shortname."""
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        else:
            self._shortname = value

    @property
    def creation_date(self):
        """:type: :class:`str`

        :getter: Returns this job's creation date

        Creation date of the job (UTC Time)
        """
        return self._creation_date

    @property
    def max_wall_time(self):
        """:type: :class:`str`
        :getter: Returns this job's maximum wall time
        :setter: Sets this job's maximum wall time

        The job's maximum wall time.
        It is a time span string.
        Format example: 'd.hh:mm:ss' or 'hh:mm:ss'

        Can be set until job is submitted.
        """
        return self._max_wall_time

    @max_wall_time.setter
    def max_wall_time(self, value):
        """Setter for maximum wall time. In time span format example : 'd.hh:mm:ss' or 'hh:mm:ss' """
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        elif _util.is_string(value):
            self._max_wall_time = value
        elif isinstance(value, datetime.timedelta):
            self._max_wall_time = _util.convert_timedelta_to_timespan_string(value)
        else:
            raise TypeError("Maximum wall time must be a time span format string (example: 'd.hh:mm:ss' or 'hh:mm:ss')")

    @property
    def pool(self):
        """:type: :class:`~qarnot.pool.Pool`
        :getter: Returns this job's pool
        :setter: Sets this job's pool

        The pool to run the job in.

        Can be set until :meth:`submit` is called.
        """
        return self._connection.retrieve_pool(self._pool_uuid)

    @pool.setter
    def pool(self, value):
        """setter for pool"""
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        else:
            self._pool_uuid = value.uuid

    @staticmethod
    def _retrieve(connection, uuid):
        """Retrieve a submitted job given its uuid.

        :param qarnot.connection.Connection connection:
          the cluster to retrieve the job from
        :param str uuid: the uuid of the job to retrieve

        :rtype: Job
        :returns: The retrieved job.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingJobException: no such job
        """
        resp = connection._get(get_url('job update', uuid=uuid))
        if resp.status_code == 404:
            raise MissingJobException(resp.json()['message'])
        raise_on_error(resp)
        return Job.from_json(connection, resp.json())

    @classmethod
    def from_json(cls, connection, payload):
        """Create a Job object from a json job.

        :param qarnot.connection.Connection connection: the cluster connection
        :param dict json_job: Dictionary representing the job
        :returns: The created :class:`~qarnot.job.Job`.
        """
        job = cls(connection,
                  payload["name"],
                  payload["poolUuid"],
                  payload["shortname"],
                  payload["useDependencies"])

        job._uuid = payload["uuid"]
        job._state = payload["state"]
        job._creation_date = payload["creationDate"]
        job._last_modified = payload["lastModified"]
        job._max_wall_time = _util.parse_timedelta(payload["maxWallTime"])

        job._auto_delete = payload["autoDeleteOnCompletion"]
        job._completion_time_to_live = payload["completionTimeToLive"]
        return job

    def _to_json(self):
        """Get a dict ready to be json packed from this task."""
        json_job = {
            'name': self._name,
            'poolUuid': self._pool_uuid,
            'shortname': self._shortname,
            'state': self._state,
            'tags': self._tags,
            'useDependencies': self._use_dependencies,
            'maxWallTime': self._max_wall_time,
            'autoDeleteOnCompletion': self._auto_delete,
            'completionTimeToLive': self._completion_time_to_live
        }

        return json_job

    def _update(self, json_job):
        """Update this job from retrieved info."""
        self._uuid = json_job['uuid']
        self._name = json_job['name']
        self._shortname = json_job.get('shortname')
        self._pool_uuid = json_job.get('poolUuid')
        self._use_dependencies = json_job.get('useDependencies')
        self._state = json_job['state']
        self._creation_date = _util.parse_datetime(json_job['creationDate'])
        self._last_modified = json_job.get('lastModified')
        self._max_wall_time = json_job.get('maxWallTime')
        self._tags = json_job.get('tags', None)
        self._previous_state = json_job.get('previousState', None)
        self._state_transition_time = json_job.get('stateTransitionTime', None)
        self._previous_state_transition_time = json_job.get('previousStateTransitionTime', None)

    def submit(self):
        """Submit job to the cluster if it is not already submitted.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.MaxJobException: Job quota reached
        :raises qarnot.exceptions.NotEnoughCreditsException: Not enough credits
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        """
        if self._uuid is not None and self._uuid != "":
            return self._state
        payload = self._to_json()
        resp = self._connection._post(get_url('jobs'), json=payload)

        if resp.status_code == 404:
            raise MissingJobException(resp.json()['message'])
        elif resp.status_code == 403:
            raise MaxJobException(resp.json()['message'])
        elif resp.status_code == 402:
            raise NotEnoughCreditsException(resp.json()['message'])
        raise_on_error(resp)
        self._uuid = resp.json()['uuid']
        self.update(True)

    def update(self, flushcache=False):
        """
        Update the job object from the REST Api.
        The flushcache parameter can be used to force the update, otherwise a cached version of the object
        will be served when accessing properties of the object.
        Cache behavior is configurable with :attr:`auto_update` and :attr:`update_cache_time`.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingJobException: job does not exist
        """
        if self._uuid is None:
            return

        now = time.time()
        if (now - self._last_cache) < self._update_cache_time and not flushcache:
            return

        resp = self._connection._get(
            get_url('job update', uuid=self._uuid))
        if resp.status_code == 404:
            raise MissingJobException(resp.json()['message'])

        raise_on_error(resp)
        self._update(resp.json())
        self._last_cache = time.time()

    def terminate(self):
        """Terminate this job on the server and abort all remaining tasks in the job.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingJobException: job does not exist
        """

        if self._uuid is None:
            return
        resp = self._connection._post(get_url('job terminate', uuid=self._uuid))
        if resp.status_code == 404:
            raise MissingJobException(resp.json()['message'])
        raise_on_error(resp)
        self._state = JobState.Terminating

    def delete(self, forceAbort=False):
        """Delete this job on the server.

        The forceAbort parameter can be used to force running task in the job to be aborted,

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.UnauthorizedException: job still contains running tasks
        :raises qarnot.exceptions.MissingJobException: job does not exist
        """

        if self._uuid is None:
            return
        resp = self._connection._delete(get_url('job delete', uuid=self._uuid, force=forceAbort))
        if resp.status_code == 404:
            raise MissingJobException(resp.json()['message'])
        elif resp.status_code == 403:
            raise UnauthorizedException(resp.json()['message'])
        raise_on_error(resp)
        self._state = JobState.Deleting
        self._uuid = None

    @property
    def auto_delete(self):
        """Autodelete this job if it is finished and your max number of job is reach

        Can be set until :meth:`submit` is called.

        :type: :class:`bool`
        :getter: Returns is this job must autodelete
        :setter: Sets this job's autodelete
        :default_value: "False"

        :raises AttributeError: if you try to reset the auto_delete after the job is submit
        """
        return self._auto_delete

    @auto_delete.setter
    def auto_delete(self, value):
        """Setter for auto_delete, this can only be set before job's submission
        """
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched job")
        self._auto_delete = value

    @property
    def completion_ttl(self):
        """The job will be auto delete `completion_ttl` after it is finished

        Can be set until :meth:`submit` is called.

        :getter:  Returns this job's completed time to live.
        :type: :class:`str`
        :setter: Sets this job's this job's completed time to live.
        :type: :class:`str` or :class:`datetime.timedelta`
        :default_value: ""

        :raises AttributeError: if you try to set it after the job is submitted

        The `completion_ttl` must be a timedelta or a time span format string (example: 'd.hh:mm:ss' or 'hh:mm:ss')
        """
        return self._completion_time_to_live

    @completion_ttl.setter
    def completion_ttl(self, value):
        """Setter for completion_ttl, this can only be set before job's submission"""
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        self._completion_time_to_live = _util.parse_to_timespan_string(value)

    @property
    def tags(self):
        """:type: :class:list(`str`)
        :getter: Returns this job's tags
        :setter: Sets this job's tags

        Custom tags.
        """
        if self._auto_update:
            self.update()

        return self._tags

    @tags.setter
    def tags(self, value):
        """Setter for tags"""
        if self._auto_update:
            self.update()
        self._tags = value

    @property
    def previous_state(self):
        """
        :type: :class:`str`
        :getter: Returns the running job's previous state
        """
        return self._previous_state

    @property
    def state_transition_time(self):
        """
        :type: :class:`str`
        :getter: Returns the running job's transition state time

        job state transition time (UTC Time)
        """
        return self._state_transition_time

    @property
    def previous_state_transition_time(self):
        """
        :type: :class:`str`
        :getter: Returns the running job's previous transition state time

        job previous state transition time (UTC Time)
        """
        return self._previous_state_transition_time

    def __repr__(self):
        return '{0} - {1} - {2} - Pool : {3} - {4} - Tags: {5} - UseDependencies : {6} '\
            .format(self.name,
                    self.shortname,
                    self._uuid,
                    self._pool_uuid,
                    self.state,
                    self._tags,
                    self._use_dependencies)



import pytest
# import mock
import qarnot
from qarnot.pool import Pool
from qarnot.task import Task
from qarnot.bucket import Bucket
from qarnot.advanced_bucket import BucketPrefixFiltering, PrefixResourcesTransformation
import datetime
from .mock_connection import MockConnection

class TestAdvancedResourceBucketsMethods:
    def test_check_the_init_values(self):
        bucket = Bucket(MockConnection(), "name", False)
        bucket2 = bucket.with_filtering(BucketPrefixFiltering("test")).with_resource_transformation(PrefixResourcesTransformation("test2"))
        assert "name" == bucket2.uuid
        assert "test" == bucket2._filtering._filters["prefixFiltering"].prefix
        assert "test2" == bucket2._resources_transformation._resource_transformers["stripPrefix"].prefix
        bucket = Bucket(MockConnection(), "name", False)
        bucket2 = bucket.with_resource_transformation(PrefixResourcesTransformation("test2")).with_filtering(BucketPrefixFiltering("test"))
        assert "name" == bucket2.uuid
        assert "test" == bucket2._filtering._filters["prefixFiltering"].prefix
        assert "test2" == bucket2._resources_transformation._resource_transformers["stripPrefix"].prefix

    def test_create_an_advance_resource_json(self):
        """
        {
            BucketName:"name",
            Filtering: {
                BucketPrefixFiltering: {
                    Prefix:"prefix"
                }
            },
            ResourcesTransformation: {
                StripPrefix: {
                    Prefix:"prefix"
                }
            }
        }
        """
        bucket = Bucket(MockConnection(), "name", False)
        bucket2 = bucket.with_filtering(BucketPrefixFiltering("prefix1")).with_resource_transformation(PrefixResourcesTransformation("prefix2"))
        json_dict = bucket2.to_json()
        assert "name" == json_dict["bucketName"]
        assert "prefix1" == json_dict["filtering"]["prefixFiltering"]["prefix"]
        assert "prefix2" == json_dict["resourcesTransformation"]["stripPrefix"]["prefix"]

    def test_create_an_advance_bucket_from_a_json(self):
        json = {
            "bucketName": "name",
            "filtering": {
                "prefixFiltering": {
                    "prefix": "prefix1"
                }
            },
            "resourcesTransformation": {
                "stripPrefix": {
                    "prefix": "prefix2"
                }
            }
        }

        bucket = Bucket.from_json(MockConnection(), json)
        assert "name" == bucket.uuid
        assert "prefix1" == bucket._filtering._filters["prefixFiltering"].prefix
        assert "prefix2" == bucket._resources_transformation._resource_transformers["stripPrefix"].prefix



"""Module to handle a pool."""

# Copyright 2017 Qarnot computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import warnings

from . import raise_on_error, get_url, _util
from .bucket import Bucket
from .status import Status
from .error import Error
from .exceptions import MissingPoolException, MaxPoolException, NotEnoughCreditsException, \
    BucketStorageUnavailableException, MissingBucketException


class Pool(object):
    """Represents a Qarnot pool.

    .. note::
       A :class:`Pool` must be created with
       :meth:`qarnot.connection.Connection.create_pool`
       or retrieved with :meth:`qarnot.connection.Connection.pools` or :meth:`qarnot.connection.Connection.retrieve_pool`.
    """

    def __init__(self, connection, name, profile, instancecount=1, shortname=None):
        """Create a new :class:`Pool`.

        :param connection: the cluster on which to send the pool
        :type connection: :class:`qarnot.connection.Connection`
        :param name: given name of the pool
        :type name: :class:`str`
        :param profile: which profile to use with this task
        :type profile: :class:`str`

        :param instancecount: number of instances or ranges on which to run pool
        :type instancecount: int or str
        :param shortname: userfriendly pool name
        :type shortname: :class:`str`
        """

        self._name = name
        self._shortname = shortname
        self._state = 'UnSubmitted'  # RO property same for below
        self._profile = profile
        self._connection = connection
        self._constants = {}
        """
         :type: dict(str, str)

         Constants of the task.
         Can be set until :meth:`submit` is called

        .. note:: See available constants for a specific profile
              with :meth:`qarnot.connection.Connection.retrieve_profile`.
        """
        self._constraints = {}
        self._auto_update = True
        self._last_auto_update_state = self._auto_update
        self._update_cache_time = 5

        self._last_cache = time.time()
        self._instancecount = instancecount
        self._resource_object_advanced = []
        self._resource_object_ids = []
        self._resource_objects = []
        self._tags = []
        self._errors = None
        self._creation_date = None
        self._uuid = None
        self._is_summary = False
        self._preparation_task = None
        self._status = None

        self._is_elastic = False
        self._elastic_minimum_slots = 0
        self._elastic_maximum_slots = 1
        self._elastic_minimum_idle_slots = 0
        self._elastic_resize_period = 90
        self._elastic_resize_factor = 1
        self._elastic_minimum_idle_time = 0
        self._running_core_count = 0
        self._running_instance_count = 0
        self._pool_usage = 0.0
        self._total_slot_capacity = 0
        self._queued_or_running_task_instances_count = 0.0

        self._completion_time_to_live = "00:00:00"
        self._auto_delete = False
        self._tasks_wait_for_synchronization = False

        self._previous_state = None
        self._state_transition_time = None
        self._previous_state_transition_time = None
        self._last_modified = None
        self._execution_time = None
        self._end_date = None

    @classmethod
    def _retrieve(cls, connection, uuid):
        """Retrieve a submitted pool given its uuid.

        :param qarnot.connection.Connection connection:
          the cluster to retrieve the pool from
        :param str uuid: the uuid of the pool to retrieve

        :rtype: Pool
        :returns: The retrieved pool.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingPoolException: no such pool
        """
        resp = connection._get(get_url('pool update', uuid=uuid))
        if resp.status_code == 404:
            raise MissingPoolException(resp.json()['message'])
        raise_on_error(resp)
        return Pool.from_json(connection, resp.json())

    @classmethod
    def from_json(cls, connection, json_pool, is_summary=False):
        """Create a Pool object from a json pool.

        :param qarnot.connection.Connection connection: the cluster connection
        :param dict json_pool: Dictionary representing the pool
        :returns: The created :class:`~qarnot.pool.Pool`.
        """
        new_pool = cls(connection,
                       json_pool['name'],
                       json_pool['profile'],
                       json_pool['instanceCount'])
        new_pool._update(json_pool)
        new_pool._is_summary = is_summary
        return new_pool

    def _update(self, json_pool):
        """Update this pool from retrieved info."""
        self._name = json_pool['name']
        self._shortname = json_pool.get('shortname')
        self._profile = json_pool['profile']
        self._instancecount = json_pool['instanceCount']

        if json_pool['runningCoreCount'] is not None:
            self._running_core_count = json_pool['runningCoreCount']
        if json_pool['runningInstanceCount'] is not None:
            self._running_instance_count = json_pool['runningInstanceCount']

        if 'errors' in json_pool:
            self._errors = [Error(d) for d in json_pool['errors']]
        else:
            self._errors = []

        if 'resourceBuckets' in json_pool and json_pool['resourceBuckets'] is not None:
            self._resource_object_ids = json_pool['resourceBuckets']

        if 'advancedResourceBuckets' in json_pool and json_pool['advancedResourceBuckets']:
            self._resource_object_advanced = json_pool['advancedResourceBuckets']

        if 'status' in json_pool:
            self._status = json_pool['status']
        self._creation_date = _util.parse_datetime(json_pool['creationDate'])

        if 'constants' in json_pool:
            for constant in json_pool['constants']:
                self._constants[constant.get('key')] = constant.get('value')
        self._uuid = json_pool['uuid']
        self._state = json_pool['state']
        self._preparation_task = json_pool.get('preparationTask')
        self._tags = json_pool.get('tags', None)
        self._tasks_wait_for_synchronization = json_pool.get('taskDefaultWaitForPoolResourcesSynchronization', False)
        self._pool_usage = json_pool.get("poolUsage", 0.0)
        self._total_slot_capacity = json_pool.get("totalSlotCapacity", 0)
        self._queued_or_running_task_instances_count = json_pool.get("queuedOrRunningTaskInstancesCount", 0.0)

        if 'autoDeleteOnCompletion' in json_pool:
            self._auto_delete = json_pool["autoDeleteOnCompletion"]
        if 'completionTimeToLive' in json_pool:
            self._completion_time_to_live = json_pool["completionTimeToLive"]

        if 'elasticProperty' in json_pool:
            elasticProperty = json_pool["elasticProperty"]
            self._is_elastic = elasticProperty["isElastic"]
            self._elastic_maximum_slots = elasticProperty["maxTotalSlots"]
            self._elastic_minimum_slots = elasticProperty["minTotalSlots"]
            self._elastic_minimum_idle_slots = elasticProperty["minIdleSlots"]
            self._elastic_minimum_idle_time = elasticProperty["minIdleTimeSeconds"]
            self._elastic_resize_factor = elasticProperty["rampResizeFactor"]
            self._elastic_resize_period = elasticProperty["resizePeriod"]

        self._previous_state = json_pool.get('previousState', None)
        self._state_transition_time = json_pool.get('stateTransitionTime', None)
        self._previous_state_transition_time = json_pool.get('previousStateTransitionTime', None)
        self._last_modified = json_pool.get('lastModified', None)
        self._execution_time = json_pool.get('executionTime', None)
        self._end_date = json_pool.get('endDate', None)

    def _to_json(self):
        """Get a dict ready to be json packed from this pool."""
        const_list = [
            {'key': key, 'value': value}
            for key, value in self._constants.items()
        ]
        constr_list = [
            {'key': key, 'value': value}
            for key, value in self._constraints.items()
        ]

        elastic_dict = {
            "isElastic": self._is_elastic,
            "minTotalSlots": self._elastic_minimum_slots,
            "maxTotalSlots": self._elastic_maximum_slots,
            "minIdleSlots": self._elastic_minimum_idle_slots,
            "resizePeriod": self._elastic_resize_period,
            "rampResizeFactor": self._elastic_resize_factor,
            "minIdleTimeSeconds": self._elastic_minimum_idle_time
        }

        json_pool = {
            'name': self._name,
            'profile': self._profile,
            'constants': const_list,
            'constraints': constr_list,
            'instanceCount': self._instancecount,
            'tags': self._tags,
            'preparationTask': self._preparation_task,
            'elasticProperty': elastic_dict,
            'taskDefaultWaitForPoolResourcesSynchronization': self._tasks_wait_for_synchronization
        }

        if self._shortname is not None:
            json_pool['shortname'] = self._shortname

        self._resource_object_advanced = [x.to_json() for x in self._resource_objects]
        json_pool['advancedResourceBuckets'] = self._resource_object_advanced

        json_pool['autoDeleteOnCompletion'] = self._auto_delete
        json_pool['completionTimeToLive'] = self._completion_time_to_live

        return json_pool

    def submit(self):
        """Submit pool to the cluster if it is not already submitted.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.MaxPoolException: Pool quota reached
        :raises qarnot.exceptions.NotEnoughCreditsException: Not enough credits
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials

        .. note:: Will ensure all added files are on the resource bucket
           regardless of their uploading mode.
        """
        if self._uuid is not None:
            return self._state
        for bucket in self.resources:
            bucket.flush()
        payload = self._to_json()
        resp = self._connection._post(get_url('pools'), json=payload)

        if resp.status_code == 404:
            raise MissingBucketException(resp.json()['message'])
        elif resp.status_code == 403:
            raise MaxPoolException(resp.json()['message'])
        elif resp.status_code == 402:
            raise NotEnoughCreditsException(resp.json()['message'])
        raise_on_error(resp)
        self._uuid = resp.json()['uuid']
        self.update(True)

    def update(self, flushcache=False):
        """
        Update the pool object from the REST Api.
        The flushcache parameter can be used to force the update, otherwise a cached version of the object
        will be served when accessing properties of the object.
        Cache behavior is configurable with :attr:`auto_update` and :attr:`update_cache_time`.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingTaskException: pool does not represent a
          valid one
        """
        if self._uuid is None:
            return

        now = time.time()
        if (now - self._last_cache) < self._update_cache_time and not flushcache:
            return

        resp = self._connection._get(
            get_url('pool update', uuid=self._uuid))
        if resp.status_code == 404:
            raise MissingPoolException(resp.json()['message'])

        raise_on_error(resp)
        self._update(resp.json())
        self._is_summary = False
        self._last_cache = time.time()

    def commit(self):
        """Replicate local changes on the current object instance to the REST API

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials

        This function need to be call to apply the local elastic pool setting modifications.
        .. note:: When updating buckets' properties, auto update will be disabled until commit is called.
        """
        response = self._connection._put(get_url('pool update', uuid=self._uuid), json=self._to_json())
        self._auto_update = self._last_auto_update_state

        if response.status_code == 404:
            raise MissingPoolException(response.json()['message'])
        raise_on_error(response)

    def update_resources(self):
        """ Update resources for a running pool.

        The typical workflow is as follows:
           1. Upload new files on your resource bucket,
           2. Call this method,
           3. The new files will appear on all the compute nodes in the $DOCKER_WORKDIR folder
        Note: There is no way to know when the files are effectively transfered. This information is available on the compute node only.
        Note: The update is additive only: files deleted from the bucket will NOT be deleted from the pool's resources directory.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingPoolException: pool does not exist
        """

        self.update(True)
        resp = self._connection._patch(get_url('pool update', uuid=self._uuid))

        if resp.status_code == 404:
            raise MissingPoolException(resp.json()['message'])
        raise_on_error(resp)

        self.update(True)

    def setup_elastic(self, minimum_total_slots=0, maximum_total_slots=1, minimum_idle_slots=0, minimum_idle_time_seconds=0, resize_factor=1, resize_period=90):
        """Setup the pool elastic properties

        :param int minimum_total_slots: Minimum slot number for the pool in elastic mode.
                Defaults to 0.
        :param int maximum_total_slots: Maximum slot number for the pool in elastic mode.
                Defaults to 1.
        :param int minimum_idle_slots: Minimum idling slot number.
                Defaults to 0.
        :param int minimum_idle_time_seconds: Wait time in seconds before closing an unused slot if the number of unused slots are upper than the minimum_idle_slots.
                Defaults to 0.
        :param float resize_factor: Growing factor of the pool. It must be a number between 0 and 1.
                Defaults to 1.
        :param int resize_period: Refresh rate of resizing the pool in elastic mode.
                Defaults to 90.
        """
        self._is_elastic = True
        self._elastic_maximum_slots = maximum_total_slots
        self._elastic_minimum_slots = minimum_total_slots
        self._elastic_minimum_idle_slots = minimum_idle_slots
        self._elastic_minimum_idle_time = minimum_idle_time_seconds
        self._elastic_resize_factor = resize_factor
        self._elastic_resize_period = resize_period

    def delete(self, purge_resources=False):
        """Delete this pool on the server.

        :param bool purge_resources: parameter value is used to determine if the bucket is also deleted.
                Defaults to False.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingTaskException: pool does not exist
        """
        if purge_resources:
            self._update_if_summary()

        if self._auto_update:
            self._auto_update = False

        if self._uuid is None:
            return

        resp = self._connection._delete(get_url('pool update', uuid=self._uuid))
        if resp.status_code == 404:
            raise MissingPoolException(resp.json()['message'])
        raise_on_error(resp)

        if purge_resources and len(self.resources) != 0:
            toremove = []
            for r in self.resources:
                try:
                    r.update()
                    r.delete()
                    toremove.append(r)
                except (MissingBucketException, BucketStorageUnavailableException) as exception:
                    warnings.warn(str(exception))
            for tr in toremove:
                self.resources.remove(tr)

        self._state = "Deleted"
        self._uuid = None

    def close(self):
        """Close this pool if running.

        :raises qarnot.exceptions.QarnotGenericException: API general error, see message for details
        :raises qarnot.exceptions.UnauthorizedException: invalid credentials
        :raises qarnot.exceptions.MissingPoolException: pool does not exist
        """
        self.update(True)

        resp = self._connection._post(
            get_url('pool close', uuid=self._uuid))

        if resp.status_code == 404:
            raise MissingPoolException(resp.json()['message'])
        raise_on_error(resp)

        self.update(True)

    @property
    def uuid(self):
        """:type: :class:`str`
        :getter: Returns this pool's uuid

        The pool's uuid.

        Automatically set when a pool is submitted.
        """
        return self._uuid

    @property
    def state(self):
        """:type: :class:`str`
        :getter: return this pool's state

        State of the pool.

        Value is in
           * UnSubmitted
           * Submitted
           * PartiallyDispatched
           * FullyDispatched
           * PartiallyExecuting
           * FullyExecuting
           * Closing
           * Closed
           * Failure
           * PendingDelete

        .. warning::
           this is the state of the pool when the object was retrieved,
           call :meth:`update` for up to date value.
        """
        if self._auto_update:
            self.update()
        return self._state

    @property
    def resources(self):
        """:type: list(:class:`~qarnot.bucket.Bucket`)
        :getter: Returns this pool's resources bucket
        :setter: Sets this pool's resources bucket

        Represents resource files.
        """
        if self._auto_update:
            self.update()
        if not self._resource_objects:
            for adv in self._resource_object_advanced:
                d = Bucket.from_json(self._connection, adv)
                self._resource_objects.append(d)

            for bid in self._resource_object_ids:
                d = Bucket(self._connection, bid)
                self._resource_objects.append(d)

        return self._resource_objects

    @resources.setter
    def resources(self, value):
        """This is a setter."""
        self._resource_objects = value

    @property
    def name(self):
        """:type: :class:`str`
        :getter: Returns this pool's name
        :setter: Sets this pool's name

        The pool's name.

        Can be set until pool is submitted.
        """
        return self._name

    @name.setter
    def name(self, value):
        """Setter for name."""
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        else:
            self._name = value

    @property
    def shortname(self):
        """:type: :class:`str`
        :getter: Returns this pool's shortname
        :setter: Sets this pool's shortname

        The pool's shortname, must be DNS compliant and unique, if not provided, will default to :attr:`uuid`.

        Can be set until pool is submitted.
        """
        return self._shortname

    @shortname.setter
    def shortname(self, value):
        """Setter for shortname."""
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        else:
            self._shortname = value

    @property
    def tags(self):
        """:type: :class:list(`str`)
        :getter: Returns this pool's tags
        :setter: Sets this pool's tags

        Custom tags.
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        return self._tags

    @tags.setter
    def tags(self, value):
        """Setter for tags"""
        self._tags = value
        self._auto_update = False

    @property
    def profile(self):
        """:type: :class:`str`
        :getter: Returns this pool's profile
        :setter: Sets this pool's profile

        The profile to run the pool with.

        Can be set until :meth:`submit` is called.
        """
        return self._profile

    @profile.setter
    def profile(self, value):
        """setter for profile"""
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        else:
            self._profile = value

    @property
    def instancecount(self):
        """:type: :class:`int`
        :getter: Returns this pool's instance count
        :setter: Sets this pool's instance count

        Number of instances needed for the pool.

        Can be set until :meth:`submit` is called.
        """
        return self._instancecount

    @instancecount.setter
    def instancecount(self, value):
        """Setter for instancecount."""
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        self._instancecount = value

    @property
    def running_core_count(self):
        """:type: :class:`int`
        :getter: Returns this pool's running core count

        Number of core running inside the pool.
        """
        return self._running_core_count

    @property
    def running_instance_count(self):
        """:type: :class:`int`
        :getter: Returns this pool's running instance count

        Number of instances running inside the pool.
        """
        return self._running_instance_count

    @property
    def errors(self):
        """:type: list(:class:`str`)
        :getter: Returns this pool's error list

        Error reason if any, empty string if none
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        return self._errors

    @property
    def creation_date(self):
        """:type: :class:`str`

        :getter: Returns this pool's creation date

        Creation date of the pool (UTC Time)
        """
        return self._creation_date

    @property
    def status(self):
        """:type: :class:`qarnot.status.Status`
        :getter: Returns this pool's status

        Status of the pool
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        if self._status:
            return Status(self._status)
        return self._status

    @property
    def auto_update(self):
        """:type: :class:`bool`

        :getter: Returns this pool's auto update state
        :setter: Sets this pool's auto update state

        Auto update state, default to True
        When auto update is disabled properties will always return cached value
        for the object and a call to :meth:`update` will be required to get latest values from the REST Api.
        """
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value):
        """Setter for auto_update feature
        """
        self._auto_update = value
        self._last_auto_update_state = self._auto_update

    @property
    def update_cache_time(self):
        """:type: :class:`int`

        :getter: Returns this pool's auto update state
        :setter: Sets this pool's auto update state

        Cache expiration time, default to 5s
        """
        return self._update_cache_time

    @update_cache_time.setter
    def update_cache_time(self, value):
        """Setter for update_cache_time
        """
        self._update_cache_time = value

    def _update_if_summary(self):
        """Trigger flush update if the task is made from a summary.

        This should be called before accessing any fields not contained in a summary task
        """
        if self._is_summary:
            self.update(True)

    @property
    def is_elastic(self):
        """:type: :class:`bool`
        :getter: Returns this pool's is_elastic
        :setter: Sets this pool's is_elastic

        Define if you use a static or an elastic pool.
        """
        return self._is_elastic

    @is_elastic.setter
    def is_elastic(self, value):
        self._is_elastic = value

    @property
    def elastic_minimum_slots(self):
        """:type: :class:`int`
        :getter: Returns this pool's elastic_minimum_slots
        :setter: Sets this pool's elastic_minimum_slots

        The minimum slot number of the elastic pool.
        Define the minimum number of pool instances stay open during the pool execution.
        """
        return self._elastic_minimum_slots

    @elastic_minimum_slots.setter
    def elastic_minimum_slots(self, value):
        """Setter for elastic_minimum_slots"""
        self._elastic_minimum_slots = value

    @property
    def elastic_maximum_slots(self):
        """:type: :class:`int`
        :getter: Returns this pool's elastic_maximum_slots
        :setter: Sets this pool's elastic_maximum_slots

        The maximum slot number of the elastic pool.
        Define the maximum number of pool instances opened during the pool execution.
        """
        return self._elastic_maximum_slots

    @elastic_maximum_slots.setter
    def elastic_maximum_slots(self, value):
        """Setter for elastic_maximum_slots"""
        self._elastic_maximum_slots = value

    @property
    def elastic_minimum_idle_slots(self):
        """:type: :class:`int`
        :getter: Returns this pool's elastic_minimum_idle_slots
        :setter: Sets this pool's elastic_minimum_idle_slots

        The minimum idle number of the elastic pool.
        Define the minimum number of the idle pool instances stay opened during the pool execution.
        It should be lower to elastic_minimum_slots to be usefull
        """
        return self._elastic_minimum_idle_slots

    @elastic_minimum_idle_slots.setter
    def elastic_minimum_idle_slots(self, value):
        """Setter for elastic_minimum_idle_slots"""
        self._elastic_minimum_idle_slots = value

    @property
    def elastic_minimum_idle_time(self):
        """:type: :class:`int`
        :getter: Returns this pool's elastic_minimum_idle_time
        :setter: Sets this pool's elastic_minimum_idle_time

        Wait time in seconds before closing an unused slot if the number of unused slots are upper than the minimum_idle_slots.
        """
        return self._elastic_minimum_idle_time

    @elastic_minimum_idle_time.setter
    def elastic_minimum_idle_time(self, value):
        """Setter for elastic_minimum_idle_time"""
        self._elastic_minimum_idle_time = value

    @property
    def elastic_resize_factor(self):
        """:type: :class:`float`
        :getter: Returns this pool's elastic_resize_factor
        :setter: Sets this pool's elastic_resize_factor

        The resize factor of the pool.
        It represent the resize factor of the slots.
        It's a decimal number upper than 0 and and equal or lower the 1
        """
        return self._elastic_resize_factor

    @elastic_resize_factor.setter
    def elastic_resize_factor(self, value):
        """Setter for elastic_resize_factor

        :raises Exception: resize factor must be > 0
        :raises Exception: resize factor must be <= 1
        """
        if value <= 0:
            raise Exception("resize factor must be > 0")
        elif value > 1:
            raise Exception("resize factor must be <= 1")
        self._elastic_resize_factor = value

    @property
    def elastic_resize_period(self):
        """:type: :class:`int`
        :getter: Returns this pool's elastic_resize_period
        :setter: Sets this pool's elastic_resize_period

        The resize period of the elastic pool in second.
        This is the refresh rate of resizing the elastic pool.
        """
        return self._elastic_resize_period

    @elastic_resize_period.setter
    def elastic_resize_period(self, value):
        """Setter for elastic_resize_period"""
        self._elastic_resize_period = value

    @property
    def preparation_command_line(self):
        """:type: :class:`str`:
        :getter: Returns this pool's command line.
        :setter: set the pool's command line.

        Update the pool command line if needed
        The command line is a command executed on the node before any task is executed.
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        if self._preparation_task:
            return self._preparation_task["commandLine"]
        return None

    @preparation_command_line.setter
    def preparation_command_line(self, value):
        """Setter for command line
        """
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        self._update_if_summary()
        if self._auto_update:
            self.update()
        if self._preparation_task:
            self._preparation_task["commandLine"] = value
        else:
            self._preparation_task = {"commandLine": value}

    @property
    def constants(self):
        """:type: dictionary{:class:`str` : :class:`str`}
        :getter: Returns this pool's constants dictionary.
        :setter: set the pool's constants dictionary.

        Update the constants if needed
        Constants are the parametrazer of the profils.
        Use them to adjust your profile parametter.
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        return self._constants

    @constants.setter
    def constants(self, value):
        """Setter for constants
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        self._constants = value

    @property
    def constraints(self):
        """:type: dictionary{:class:`str` : :class:`str`}
        :getter: Returns this pool's constraints dictionary.
        :setter: set the pool's constraints dictionary.

        Update the constraints if needed
        advance usage
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        return self._constraints

    @constraints.setter
    def constraints(self, value):
        """Setter for constraints
        """
        self._update_if_summary()
        if self._auto_update:
            self.update()

        self._constraints = value

    @property
    def tasks_default_wait_for_pool_resources_synchronization(self):
        """:type: :class:`bool`
        :getter: Returns this task's tasks_default_wait_for_pool_resources_synchronization.
        :setter: set the task's tasks_default_wait_for_pool_resources_synchronization.

        :raises qarnot.exceptions.AttributeError: can't set this attribute on a launched task
        """
        return self._tasks_wait_for_synchronization

    @tasks_default_wait_for_pool_resources_synchronization.setter
    def tasks_default_wait_for_pool_resources_synchronization(self, value):
        """Setter for tasks_default_wait_for_pool_resources_synchronization
        """
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched task")
        self._tasks_wait_for_synchronization = value

    @property
    def auto_delete(self):
        """Autodelete this pool if it is finished and your max number of pool is reach

        Can be set until :meth:`submit` is called.

        :type: :class:`bool`
        :getter: Returns is this pool must autodelete
        :setter: Sets this pool's autodelete
        :default_value: "False"

        :raises AttributeError: if you try to reset the auto_delete after the pool is submit
        """
        self._update_if_summary()
        return self._auto_delete

    @auto_delete.setter
    def auto_delete(self, value):
        """Setter for auto_delete, this can only be set before pool's submission
        """
        self._update_if_summary()
        if self.uuid is not None:
            raise AttributeError("can't set attribute on a launched pool")
        self._auto_delete = value

    @property
    def completion_ttl(self):
        """The pool will be auto delete `completion_ttl` after it is finished

        Can be set until :meth:`submit` is called.

        :getter:  Returns this pool's completed time to live.
        :type: :class:`str`
        :setter: Sets this pool's this pool's completed time to live.
        :type: :class:`str` or :class:`datetime.timedelta`
        :default_value: ""

        :raises AttributeError: if you try to set it after the pool is submitted

        The `completion_ttl` must be a timedelta or a time span format string (example: 'd.hh:mm:ss' or 'hh:mm:ss')
        """
        self._update_if_summary()
        return self._completion_time_to_live

    @completion_ttl.setter
    def completion_ttl(self, value):
        """Setter for completion_ttl, this can only be set before pool's submission"""
        self._update_if_summary()
        if self._uuid is not None:
            raise AttributeError("can't set attribute on a submitted job")
        self._completion_time_to_live = _util.parse_to_timespan_string(value)

    @property
    def previous_state(self):
        """
        :type: :class:`str`
        :getter: Returns the running pool's previous state
        """
        self._update_if_summary()
        return self._previous_state

    @property
    def state_transition_time(self):
        """
        :type: :class:`str`
        :getter: Returns the running pool's transition state time

        pool state transition time (UTC Time)
        """
        self._update_if_summary()
        return self._state_transition_time

    @property
    def previous_state_transition_time(self):
        """
        :type: :class:`str`
        :getter: Returns the running pool's previous transition state time

        pool previous state transition time (UTC Time)
        """
        self._update_if_summary()
        return self._previous_state_transition_time

    @property
    def last_modified(self):
        """
        :type: :class:`str`
        :getter: Returns the running pool's last modification time

        pool's last modified time (UTC Time)
        """
        self._update_if_summary()
        return self._last_modified

    @property
    def execution_time(self):
        """
        :type: :class:`str`
        :getter: Returns the running pool's total CPU execution time.

        pool's execution time of all it's instances.
        """
        self._update_if_summary()
        return self._execution_time

    @property
    def end_date(self):
        """
        :type: :class:`str`
        :getter: Returns the finished pool's end date.

        pool's end date (UTC Time)
        """
        self._update_if_summary()
        return self._end_date

    @property
    def pool_usage(self):
        """
        :type: :class:`float`
        :getter: Returns the pool usage
        """
        self._update_if_summary()
        return self._pool_usage

    @property
    def total_slot_capacity(self):
        """
        :type: :class:`int`
        :getter: Returns the pool slot capacity
        """
        self._update_if_summary()
        return self._total_slot_capacity

    @property
    def queued_or_running_task_instances_count(self):
        """
        :type: :class:`int`
        :getter: Returns count of task instances dispatched in the pool
        """
        return self._queued_or_running_task_instances_count

    def __repr__(self):
        return '{0} - {1} - {2} - {3} - {5} - InstanceCount : {4} - Resources : {6} '\
            'Tag {7} - IsElastic {8} - ElasticMin {9} - ElasticMax {10} - ElasticMinIdle {11} -'\
            ' ElasticResizePeriod {12} - ElasticResizeFactor {13} - ElasticMinIdleTimeSeconds {14} - '\
            'Errors {15}'\
            .format(self.name,
                    self.shortname,
                    self._uuid,
                    self._profile,
                    self._instancecount,
                    self.state,
                    ([bucket._uuid for bucket in self._resource_objects] if self._resource_objects is not None else ""),
                    self._tags,
                    self._is_elastic,
                    self._elastic_minimum_slots,
                    self._elastic_maximum_slots,
                    self._elastic_minimum_idle_slots,
                    self._elastic_resize_period,
                    self._elastic_resize_factor,
                    self._elastic_minimum_idle_time,
                    self._errors)



import pytest
import qarnot
from qarnot.job import Job
from qarnot.pool import Pool
import datetime
from .mock_job import default_json_job
from .mock_connection import MockConnection

class TestJobProperties:
    conn = MockConnection()
    def submit_job(self, job):
        job._uuid = "submitted"

    def test_max_wall_time_accept_delta_before_submit(self):
        job = Job(self.conn, "job-name")
        delta = datetime.timedelta(days=2, hours=33, minutes=66, seconds=66)
        job.max_wall_time = delta
        assert "3.10:07:06" == job.max_wall_time

    def test_max_wall_time_accept_string_before_submit(self):
        job = Job(self.conn, "job-name")
        job.max_wall_time = "3.10:07:04"
        assert "3.10:07:04" == job.max_wall_time

    def test_max_wall_time_throw_exception_delta_after_submit(self):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        delta = datetime.timedelta(days=2, hours=33, minutes=66, seconds=66)
        with pytest.raises(AttributeError):
            job.max_wall_time = delta
        with pytest.raises(AttributeError):
            job.max_wall_time = "3.10:07:04"

    def test_max_wall_time_setting_throw_exception_with_number(self):
        job = Job(self.conn, "job-name")
        with pytest.raises(TypeError):
            job.max_wall_time = 10

    def test_name_is_set_before_submit(self):
        job = Job(self.conn, "job-name")
        job.name = "name"
        assert "name" == job.name

    def test_name_setting_throw_exception_after_submitted(self):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        with pytest.raises(AttributeError):
            job.name = "name"

    def test_shortname_is_set_before_submit(self):
        job = Job(self.conn, "job-name")
        job.shortname = "shortname"
        assert "shortname" == job.shortname

    def test_shortname_setting_throw_exception_after_submitted(self):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        with pytest.raises(AttributeError):
            job.shortname = "shortname"

    def test_use_dependencies_is_set_before_submit(self):
        job = Job(self.conn, "job-name")
        job.use_dependencies = True
        assert True == job.use_dependencies

    def test_use_dependencies_setting_throw_exception_after_submitted(self):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        with pytest.raises(AttributeError):
            job.use_dependencies = True

    def test_pool_is_set_before_submit(self):
        job = Job(self.conn, "job-name")
        pool = Pool(self.conn, "pool-name", "profile", 2, "shortname")
        job.pool = pool
        assert pool.shortname == job.pool.shortname

    def test_pool_setting_throw_exception_after_submitted(self):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        pool = Pool(self.conn, "pool-name", "profile", 2, "shortname")
        with pytest.raises(AttributeError):
            job.pool = pool

    def test_job_autodelete_default_value(self):
        job = Job(self.conn, "job-name")
        assert False == job.auto_delete

    def test_job_completion_ttl_default_value(self):
        job = Job(self.conn, "job-name")
        assert "00:00:00" == job.completion_ttl

    def test_job_autodelete_set_get(self):
        job = Job(self.conn, "job-name")
        job.auto_delete = False
        assert False == job.auto_delete
        job.auto_delete = True
        assert True == job.auto_delete

    def test_job_completion_ttl_set_get(self):
        job = Job(self.conn, "job-name")
        job.completion_ttl = datetime.timedelta(days=2, hours=33, minutes=66, seconds=66)
        assert "3.10:07:06" == job.completion_ttl
        job.completion_ttl = "4.11:08:06"
        assert "4.11:08:06" == job.completion_ttl

    def test_job_are_in_job_to_json(self):
        job = Job(self.conn, "job-name")
        job.completion_ttl = "4.11:08:06"
        job.auto_delete = True
        json_job = job._to_json()

        assert json_job['completionTimeToLive'] == '4.11:08:06'
        assert json_job['autoDeleteOnCompletion'] == True

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", None),
        ("state_transition_time", None),
        ("previous_state_transition_time", None),
    ])
    def test_job_property_default_value(self, property_name,  expected_value):
        job = Job(self.conn, "job-name")
        assert getattr(job, property_name) is expected_value

    @pytest.mark.parametrize("property_name, set_value, expected_value", [
        ("max_wall_time", datetime.timedelta(days=2, hours=33, minutes=66, seconds=66), "3.10:07:06")
    ])
    def test_job_set_property_value(self, property_name, set_value, expected_value):
        job = Job(self.conn, "job-name")
        setattr(job, property_name, set_value)
        assert getattr(job, property_name) == expected_value

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("max_wall_time", 10, TypeError)
    ])
    def test_job_set_forbidden_property_raise_exception(self, property_name, set_value, exception):
        job = Job(self.conn, "job-name")
        with pytest.raises(exception):
            setattr(job, property_name, set_value)

    @pytest.mark.parametrize("property_name, set_value, exception", [
        ("max_wall_time", datetime.timedelta(seconds=66), AttributeError),
        ("name", "test", AttributeError),
    ])
    def test_job_set_property_raise_exception_after_submitted(self, property_name, set_value, exception):
        job = Job(self.conn, "job-name")
        self.submit_job(job)
        with pytest.raises(exception):
            setattr(job, property_name, set_value)

    @pytest.mark.parametrize("property_name, expected_value", [
        ("previous_state", default_json_job["previousState"]),
        ("state_transition_time", default_json_job["stateTransitionTime"]),
        ("previous_state_transition_time", default_json_job["previousStateTransitionTime"]),
    ])
    def test_job_property_update_value(self, property_name, expected_value):
        job = Job(self.conn, "job-name")
        job._update(default_json_job)
        assert getattr(job, property_name) is expected_value

    @pytest.mark.parametrize("property_name, expected_value", [
        ("name", default_json_job["name"]),
    ])
    def test_job_property_send_to_json_representation(self, property_name, expected_value):
        job = Job(self.conn, "job-name")
        job._update(default_json_job)
        job_json = job._to_json()
        assert job_json[property_name] is expected_value
